from pathlib import Path
from typing import Annotated, Optional

from rich import print as rprint
from typer import Exit, Option, Typer

from syftbox.__version__ import __version__
from syftbox.app.cli import app as app_cli
from syftbox.client.cli import app as client_cli
from syftbox.server.cli import app as server_cli

app = Typer(
    name="SyftBox",
    help="SyftBox CLI",
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)

CONFIG_OPTS = Option("-c", "--config", "--config_path", help="Path to the SyftBox config")


@app.command(rich_help_panel="General Options")
def version():
    """Print SyftBox version"""

    print(__version__)


@app.command(rich_help_panel="General Options")
def debug(config_path: Annotated[Optional[Path], CONFIG_OPTS] = None):
    """Print SyftBox debug data"""

    # lazy import to improve CLI startup performance
    from syftbox.lib.debug import debug_report_yaml

    try:
        rprint(debug_report_yaml(config_path))
    except Exception as e:
        rprint(f"[red]Error[/red]: {e}")
        raise Exit(1)


app.add_typer(client_cli, name="client")
app.add_typer(server_cli, name="server")
app.add_typer(app_cli, name="app")


def main():
    app()


if __name__ == "__main__":
    main()



import datetime
import sys
from platform import platform

import httpx
from pydantic import BaseModel, Field
from typing_extensions import Optional

from syftbox.__version__ import __version__
from syftbox.client.env import syftbox_env
from syftbox.lib.client_config import SyftClientConfig


class ErrorReport(BaseModel):
    client_config: SyftClientConfig
    server_syftbox_version: Optional[str] = None
    client_syftbox_version: str = __version__
    python_version: str = sys.version
    platform: str = platform()
    timestamp: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    env: dict = Field(default=syftbox_env.model_dump())

    @classmethod
    def from_client_config(cls, client_config: SyftClientConfig):
        return cls(
            client_config=client_config,
            server_version=try_get_server_version(client_config.server_url),
        )


def make_error_report(client_config: SyftClientConfig):
    return ErrorReport.from_client_config(client_config)


def try_get_server_version(server_url):
    try:
        # do not use the server_client here, as it may not be in bad state
        return httpx.get(f"{server_url}/info").json()["version"]
    except Exception:
        return None



from types import SimpleNamespace

import pytest

from syftbox.app.install import check_os_compatibility


def test_os_compatibility_compatible():
    app_config_mock = SimpleNamespace(
        **{
            "app": SimpleNamespace(
                **{
                    "platforms": ["darwin", "linux"],
                }
            ),
        }
    )

    check_os_compatibility(app_config_mock)


def test_os_compatibility_incompatible():
    app_config_mock = SimpleNamespace(
        **{
            "app": SimpleNamespace(
                **{
                    "platforms": ["different_os"],
                }
            ),
        }
    )
    with pytest.raises(OSError) as e:
        check_os_compatibility(app_config_mock)
        assert e.value == "Your OS isn't supported by this app."


def test_os_compatibility_without_config():
    app_config_mock = SimpleNamespace(**{"app": {}})

    check_os_compatibility(app_config_mock)



import asyncio
import json
import secrets
import shutil
from pathlib import Path

import pytest
from loguru import logger

from tests.e2e.conftest import Client, E2EContext, Server

AGGREGATOR_CONFIG = {
    "participants": ["user1@openmined.org", "user2@openmined.org", "user3@openmined.org"],
}

AGGREGATOR_API_NAME = "model_aggregator"
LOCAL_TRAINING_API_NAME = "model_local_training"
MAX_COPY_DATA_PARTS = 2


def deployment_config():
    return {
        "e2e_name": "aggregator_with_local_training",
        "server": Server(port=5001),
        "clients": [
            Client(
                name="agg",
                port=8080,
                apps=["https://github.com/OpenMined/model_aggregator"],
            ),
            Client(
                name="user1",
                port=8081,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/model_local_training",
                ],
            ),
            Client(
                name="user2",
                port=8082,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/model_local_training",
                ],
            ),
            Client(
                name="user3",
                port=8083,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/model_local_training",
                ],
            ),
        ],
    }


async def copy_train_data_to_private(e2e_context: E2EContext, clients: list[Client]) -> dict[str, list[str]]:
    client_data_map: dict[str, list[str]] = {}

    for client in clients:
        await e2e_context.wait_for_api(LOCAL_TRAINING_API_NAME, client)
        mnist_samples_dir = client.api_path(LOCAL_TRAINING_API_NAME) / "mnist_samples"
        all_mnist_samples = list(mnist_samples_dir.glob("*.pt"))
        client_private_dir = client.private_dir / LOCAL_TRAINING_API_NAME
        client_private_dir.mkdir(parents=True, exist_ok=True)

        client_sample_files = []
        for _ in range(MAX_COPY_DATA_PARTS):
            # random.random is not concurrency safe, using secrets.randbelow
            rand_idx = secrets.randbelow(len(all_mnist_samples))
            random_mnist_sample = all_mnist_samples[rand_idx]
            logger.debug(f"Copying {random_mnist_sample} to {client.email} private dir")
            shutil.copy(random_mnist_sample, client_private_dir)
            assert Path(client_private_dir, random_mnist_sample.name).exists()
            client_sample_files.append(random_mnist_sample.name)

        client_data_map[client.email] = client_sample_files

    return client_data_map


async def wait_for_public_trained_models(e2e_context: E2EContext, client: Client, mnist_samples: list[str]):
    await e2e_context.wait_for_api(LOCAL_TRAINING_API_NAME, client)
    public_dir = client.public_dir
    for mnist_sample in mnist_samples:
        model_file = public_dir / f"trained_{mnist_sample}"
        await e2e_context.wait_for_path(model_file, timeout=240, interval=1)
        assert model_file.exists()


@pytest.mark.asyncio
@pytest.mark.parametrize("e2e_context", [deployment_config()], indirect=True, ids=["aggregator_with_local_training"])
async def test_e2e_aggregator_with_local_training(e2e_context: E2EContext):
    # setting up: Run 4 clients (1 agg + 3 participants) and the cache server, also installs the apps on the clients
    logger.info(f"Starting E2E '{e2e_context.e2e_name}'")
    e2e_context.reset_test_dir()
    agg_client = e2e_context.clients[0]

    await e2e_context.start_all()
    await e2e_context.wait_for_api(AGGREGATOR_API_NAME, agg_client)

    logger.info("Aggregator copies `participants.json` into the `launch` directory")
    launch_dir = agg_client.api_data_dir(AGGREGATOR_API_NAME) / "launch"
    launch_dir.mkdir(parents=True, exist_ok=True)
    participants_file = launch_dir / "participants.json"
    participants_file.write_text(json.dumps(AGGREGATOR_CONFIG))

    logger.info("Aggregator copies test data to the private folder")
    agg_private_dir = agg_client.private_dir / AGGREGATOR_API_NAME
    agg_private_dir.mkdir(parents=True, exist_ok=True)
    test_dataset_path = agg_client.api_path(AGGREGATOR_API_NAME) / "samples" / "test_data" / "mnist_dataset.pt"
    shutil.copy(test_dataset_path, agg_private_dir)

    clients: list[Client] = e2e_context.clients[1:]
    logger.info("Participants moving the MNIST data parts into private/model_local_training to train")
    client_data_map = await copy_train_data_to_private(e2e_context, clients)

    logger.info("Waiting for local clients to train their models")
    await asyncio.gather(
        *[
            wait_for_public_trained_models(e2e_context, client, client_data_map[client.email])
            for client in e2e_context.clients[1:]
        ]
    )

    logger.info("Waiting for aggregator to aggregate the models and generate results")
    done_dir = agg_client.api_data_dir(AGGREGATOR_API_NAME) / "done"
    results_file = done_dir / "results.json"
    await e2e_context.wait_for_path(results_file, timeout=90, interval=1)

    result = json.loads(results_file.read_text())
    logger.info(f"Validating results\n{result}")
    assert result["accuracy"] >= 10.0
    assert set(result["participants"]) == set(AGGREGATOR_CONFIG["participants"])
    assert len(result["missing_peers"]) == 0

    logger.info("Check that the launch and running dir are empty")
    assert len(list(launch_dir.iterdir())) == 0
    agg_running_dir = agg_client.api_data_dir(AGGREGATOR_API_NAME) / "running"
    assert len(list(agg_running_dir.iterdir())) == 0



import socket


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("", port))
            return False
        except socket.error:
            return True


def get_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]



from typer.testing import CliRunner

from syftbox.main import app as main_cli

# Initialize test runner
runner = CliRunner()


def test_version():
    result = runner.invoke(main_cli, ["version"])
    assert result.exit_code == 0
    assert len(result.stdout.split(".")) == 3


def test_debug(monkeypatch):
    def mock_debug_report(*args, **kwargs):
        return "ok: true"

    monkeypatch.setattr("syftbox.lib.debug.debug_report_yaml", mock_debug_report)
    result = runner.invoke(main_cli, ["debug"])
    assert result.exit_code == 0, result.stdout
    assert "ok: true" in result.stdout


def test_debug_invalid_config(monkeypatch):
    def mock_debug_report(*args, **kwargs):
        raise Exception("Invalid config")

    monkeypatch.setattr("syftbox.lib.debug.debug_report_yaml", mock_debug_report)

    result = runner.invoke(main_cli, ["debug"])
    assert result.exit_code == 1
    assert "Error" in result.stdout


def test_debug_custom_config(monkeypatch, mock_config):
    def mock_debug_report(conf_path):
        assert str(conf_path) == str(mock_config.path)
        return "ok: true"

    monkeypatch.setattr("syftbox.lib.debug.debug_report_yaml", mock_debug_report)

    result = runner.invoke(main_cli, ["debug", "--config", str(mock_config.path)])
    assert result.exit_code == 0
    assert "ok: true" in result.stdout



from typing import Union

from pydantic import BaseModel, EmailStr, NameEmail

from .constants import FROM_EMAIl


class SendEmailRequest(BaseModel):
    to: Union[EmailStr, NameEmail]
    subject: str
    html: str

    def json_for_request(self):
        return {
            "from": FROM_EMAIl,
            "to": [self.to],
            "subject": self.subject,
            "html": self.html,
        }



from pathlib import Path

import pytest

from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.client_shim import Client


@pytest.fixture
def client(mock_config):
    return Client(conf=mock_config)


def test_client_properties(client, mock_config):
    assert client.email == mock_config.email
    assert client.config_path == mock_config.path
    assert client.my_datasite == client.workspace.datasites / mock_config.email
    assert client.datasites == client.workspace.datasites
    assert client.sync_folder == client.workspace.datasites  # Test deprecated property
    assert client.datasite_path == client.workspace.datasites / mock_config.email  # Test deprecated property


def test_client_load(mock_config, tmp_path):
    _ = mock_config  # this fixture creates and saves a mock tmp_path/config.json file
    config_path = tmp_path / "config.json"
    client = Client.load(config_path)
    assert isinstance(client, Client)
    assert isinstance(client.config, SyftClientConfig)


def test_api_request_name(client, monkeypatch):
    test_dir = Path("/fake/api/request/path")
    monkeypatch.setattr(Path, "cwd", lambda: test_dir)
    assert client.api_request_name == "path"


@pytest.mark.parametrize(
    "api_name,datasite,expected",
    [
        ("test_api", None, "test_api"),
        (None, "other@example.com", "current_api"),
        ("custom_api", "other@example.com", "custom_api"),
    ],
)
def test_api_data(client, monkeypatch, api_name, datasite, expected):
    monkeypatch.setattr(Path, "cwd", lambda: Path("/fake/current_api"))

    result = client.api_data(api_name, datasite)

    datasite_email = datasite or client.config.email
    expected_path = client.workspace.datasites / datasite_email / "api_data" / (expected)
    assert result == expected_path


def test_makedirs(client, tmp_path):
    test_paths = [tmp_path / "dir1", tmp_path / "dir2" / "subdir", tmp_path / "dir3" / "subdir" / "subsubdir"]

    client.makedirs(*test_paths)

    for path in test_paths:
        assert path.exists()
        assert path.is_dir()


def test_makedirs_existing(client, tmp_path):
    test_path = tmp_path / "existing"
    test_path.mkdir()

    # Should not raise error when directory exists
    client.makedirs(test_path)
    assert test_path.exists()



import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from syftbox.app.install import InstallResult, install
from syftbox.lib.workspace import SyftWorkspace


@dataclass
class InstalledApps:
    apps_dir: Path
    apps: List[Path]


def install_app(workspace: SyftWorkspace, repository: str, branch: str = "main") -> InstallResult:
    return install(workspace.apps, repository, branch)


def list_app(workspace: SyftWorkspace) -> InstalledApps:
    apps = []
    if workspace.apps.exists() and workspace.apps.is_dir():
        apps = sorted([app for app in workspace.apps.iterdir() if app.is_dir()])
    return InstalledApps(workspace.apps, apps)


def uninstall_app(app_name: str, workspace: SyftWorkspace) -> Optional[Path]:
    app_dir = Path(workspace.apps, app_name)
    # first check for symlink
    if app_dir.exists() and app_dir.is_symlink():
        app_dir.unlink()
    elif app_dir.exists() and app_dir.is_dir():
        shutil.rmtree(app_dir)
    else:
        app_dir = None
    return app_dir


def update_app(ws: SyftWorkspace) -> None:
    pass



from syftbox.lib.types import PathLike, to_path


class SyftWorkspace:
    """
    A Syft workspace is a directory structure for everything stored by the client.
    Each workspace is expected to be unique for a client.

    ```txt
        data_dir/
        ├── apis/                       <-- installed apis
        ├── plugins/                    <-- plugins data
        └── datasites/                  <-- synced datasites
            ├── user1@openmined.org/
            │   └── api_data/
            └── user2@openmined.org/
                └── api_data/
    ```
    """

    def __init__(self, data_dir: PathLike):
        self.data_dir = to_path(data_dir)
        """Path to the root directory of the workspace."""

        # datasites dir
        self.datasites = self.data_dir / "datasites"
        """Path to the directory containing datasites."""

        # plugins dir
        """Path to the directory containing plugins."""
        self.plugins = self.data_dir / "plugins"

        # apps/apis dir
        self.apps = self.data_dir / "apis"
        """Path to the directory containing apps."""

    def mkdirs(self):
        self.datasites.mkdir(parents=True, exist_ok=True)
        self.plugins.mkdir(parents=True, exist_ok=True)
        self.apps.mkdir(parents=True, exist_ok=True)



import threading
import time
from pathlib import Path
from secrets import token_hex

from syftbox.client.plugins.apps import run_apps
from tests.integration.app_plugin.fixtures.app_mocks import AppMockFactory


def verify_app_execution(app_dir: Path, expected_output: str):
    """Verify app execution results."""
    app_log_path = app_dir / "app.log"
    assert app_log_path.exists()
    assert app_log_path.read_text().strip() == expected_output


def verify_running_apps(running_apps: dict, expected_app_name: str = None):
    """Verify running apps state."""
    if expected_app_name:
        assert len(running_apps) == 1
        assert expected_app_name in running_apps
    else:
        assert len(running_apps) == 0


def test_app_plugin_without_config(tmp_path, monkeypatch):
    """Test app plugin execution without configuration."""
    app_dir = tmp_path / token_hex(8)
    app_name = "test_app_without_config"
    mock_app_dir, expected_output = AppMockFactory.create_app_without_config(apps_dir=app_dir, app_name=app_name)

    assert mock_app_dir.exists()

    # Patch necessary attributes
    PATCHED_RUNNING = {}
    monkeypatch.setattr("syftbox.client.plugins.apps.DEFAULT_APPS_PATH", "")
    monkeypatch.setattr("syftbox.client.plugins.apps.RUNNING_APPS", PATCHED_RUNNING)

    # Run app
    no_config = ""  # dummy app doesn't need SYFTBOX_CLIENT_CONFIG_PATH
    run_apps(app_dir, no_config)

    # Verify results
    verify_running_apps(PATCHED_RUNNING)
    verify_app_execution(mock_app_dir, expected_output)


def test_app_plugin_with_config(tmp_path, monkeypatch):
    """Test app plugin execution with configuration."""
    app_dir = tmp_path / token_hex(8)
    app_name = "test_app_with_config"
    mock_app_dir, expected_output = AppMockFactory.create_app_with_config(apps_dir=app_dir, app_name=app_name)

    assert mock_app_dir.exists()

    # Patch necessary attributes
    PATCHED_RUNNING = {}
    EVT = threading.Event()
    monkeypatch.setattr("syftbox.client.plugins.apps.DEFAULT_APPS_PATH", "")
    monkeypatch.setattr("syftbox.client.plugins.apps.RUNNING_APPS", PATCHED_RUNNING)
    monkeypatch.setattr("syftbox.client.plugins.apps.EVENT", EVT)

    # Run app
    no_config = ""  # dummy app doesn't need SYFTBOX_CLIENT_CONFIG_PATH
    run_apps(app_dir, no_config)
    time.sleep(2)

    # Verify results
    verify_running_apps(PATCHED_RUNNING, app_name)
    verify_app_execution(mock_app_dir, expected_output)

    # This doesn't kill the process gracefully,
    # later need to implement a graceful shutdown mechanism for apps
    if app_name in PATCHED_RUNNING:
        EVT.set()
        app_thread: threading.Thread = PATCHED_RUNNING[app_name]
        app_thread.join(timeout=1)



import pytest
from fastapi.testclient import TestClient

from syftbox.client.api import create_api
from syftbox.client.base import SyftClientInterface


class MockClient(SyftClientInterface):
    @property
    def all_datasites(self):
        return ["datasite1", "datasite2"]


@pytest.fixture
def mock_api():
    yield TestClient(create_api({}))


def test_create_api(mock_api):
    response = mock_api.get("/")
    assert response.status_code == 200


def test_version(mock_api):
    response = mock_api.get("/version")
    assert response.status_code == 200
    assert "version" in response.json()


def test_datasites():
    app = create_api(MockClient())
    mock_api = TestClient(app)
    response = mock_api.get("/datasites")
    assert response.status_code == 200
    assert response.json() == {"datasites": ["datasite1", "datasite2"]}


def test_datasites_exception(mock_api):
    response = mock_api.get("/datasites")
    assert response.status_code == 500



import os
import tempfile
from pathlib import Path

import pytest

from syftbox.client.cli_setup import prompt_data_dir, prompt_email
from syftbox.lib.constants import DEFAULT_DATA_DIR
from syftbox.lib.validators import is_valid_dir, is_valid_email


@pytest.mark.parametrize(
    "path,expected",
    [
        ("/tmp", True),
        ("./test", True),
        (".", True),
        ("..", True),
        ("~", True),
        ("", False),  # Empty path = invalid
        ("/x", False),  # unwriteable path
    ],
)
def test_is_valid_dir(path, expected):
    """Test various email formats"""
    valid, reason = is_valid_dir(path, check_empty=False, check_writable=True)
    assert valid == expected, reason


def test_empty_dir():
    # Test with temp directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Valid empty directory
        valid, reason = is_valid_dir(temp_dir)
        assert valid
        assert reason == ""

        # Non-empty directory
        with open(os.path.join(temp_dir, "test.txt"), "w") as f:
            f.write("test")

        valid, reason = is_valid_dir(temp_dir)
        assert not valid
        assert "not empty" in reason.lower()


@pytest.mark.parametrize(
    "email,expected",
    [
        ("test@example.com", True),
        ("test.name@example.com", True),
        ("test+label@example.com", True),
        ("test@sub.example.com", True),
        ("a@b.c", True),
        ("", False),  # Empty email
        ("test@", False),  # Missing domain
        ("@example.com", False),  # Missing username
        ("test@example", False),  # Mising TLD
        ("test.example.com", False),  # Missing @
        ("test@@example.com", False),  # Double @
        ("test@exam ple.com", False),  # Space
        ("test@example..com", False),  # Double dots
    ],
)
def test_email_validation(email, expected):
    """Test various email formats"""
    assert is_valid_email(email) == expected


@pytest.mark.parametrize(
    "user_input,expected",
    [
        ("", Path(DEFAULT_DATA_DIR)),
        ("./valid/path", Path("./valid/path")),
    ],
)
def test_prompt_data_dir(user_input, expected, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda *a, **k: user_input)
    monkeypatch.setattr("syftbox.client.cli_setup.is_valid_dir", lambda x: (True, ""))

    dir = prompt_data_dir()
    assert dir.absolute() == expected.absolute()


@pytest.mark.timeout(1)
def test_prompt_email(monkeypatch):
    valid_email = "test@example.com"

    monkeypatch.setattr("builtins.input", lambda *a, **k: valid_email)
    monkeypatch.setattr("syftbox.client.cli_setup.is_valid_dir", lambda x: (True, ""))

    email = prompt_email()
    assert email == valid_email



import json
import zipfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from loguru import logger
from pydantic import BaseModel

from syftbox.server.logger import analytics_logger
from syftbox.server.sync.file_store import FileStore


def to_jsonable_dict(obj: dict) -> dict:
    """
    Convert log record to a JSON serializable dictionary.
    """
    result = {}
    for key, value in obj.items():
        if isinstance(value, dict):
            result[key] = to_jsonable_dict(value)
        elif isinstance(value, BaseModel):
            result[key] = value.model_dump(mode="json")
        elif isinstance(value, datetime):
            result[key] = value.isoformat()
        elif isinstance(value, Path):
            result[key] = value.as_posix()
        elif isinstance(value, (str, int, float, bool, type(None))):
            result[key] = value
        else:
            result[key] = str(value)

    return result


def log_analytics_event(
    endpoint: str,
    email: Optional[str],
    message: str = "",
    **kwargs: Any,
) -> None:
    """
    Log an event to the analytics logger.
    """
    email = email or "anonymous"

    try:
        extra = {
            "email": email,
            "endpoint": endpoint,
            "timestamp": datetime.now(timezone.utc),
            **kwargs,
        }
        extra = to_jsonable_dict(extra)
        analytics_logger.bind(**extra).info(message)
    except Exception as e:
        logger.error(f"Failed to log event: {e}")


def log_file_change_event(
    endpoint: str,
    email: Optional[str],
    relative_path: Path,
    file_store: FileStore,
) -> None:
    """
    Log a file change event to the analytics logger.
    """
    try:
        metadata = file_store.get_metadata(relative_path)
        log_analytics_event(
            endpoint=endpoint,
            email=email,
            file_metadata=metadata,
        )
    except Exception as e:
        logger.error(f"Failed to log file change event: {e}")


def _parse_analytics_file(file_path: Path) -> list[dict]:
    if file_path.suffix == ".zip":
        with zipfile.ZipFile(file_path, "r") as zfile:
            with zfile.open(zfile.namelist()[0]) as f:
                content = f.read().decode("utf-8")
    else:
        with open(file_path, "r") as f:
            content = f.read()

    events = []
    for line in content.split("\n"):
        if not line:
            continue

        try:
            event = json.loads(line)
            event["timestamp"] = datetime.fromisoformat(event["timestamp"])
            events.append(event)
        except Exception as e:
            logger.error(f"Failed to parse event: {e}")

    return events


def parse_analytics_logs(logs_dir: Path) -> list[dict]:
    # Load current log and all archived logs
    log_files = list(logs_dir.glob("analytics.log")) + list(logs_dir.glob("analytics*.zip"))
    logger.info(f"Loading logs from: {[f.as_posix() for f in log_files]}")
    events = []
    for log_file in log_files:
        events.extend(_parse_analytics_file(log_file))

    events = sorted(events, key=lambda x: x["timestamp"])
    return events



import os
import time

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

chrome_driver_path = os.environ["CHROMEDRIVER_PATH"]
email = os.environ["NETFLIX_EMAIL"]
password = os.environ["NETFLIX_PASSWORD"]
profile = os.environ["NETFLIX_PROFILE"]
output_dir = os.environ["OUTPUT_DIR"]

print(f"🍿 Downloading Netflix Activity for: {email} Profile {profile}")

# Set up WebDriver (for Chrome)
chrome_options = Options()
prefs = {
    "download.default_directory": output_dir,
    "download.prompt_for_download": False,
}
chrome_options.add_experimental_option("prefs", prefs)
chrome_options.add_argument("--headless")  # Run in headless mode, comment this if you want to see the browser window
chrome_service = Service(chrome_driver_path)  # Set the path to your ChromeDriver

driver = webdriver.Chrome(service=chrome_service, options=chrome_options)

# get login page
driver.get("https://www.netflix.com/login")


# Find the email and password input fields
email_input = driver.find_element(By.NAME, "userLoginId")
password_input = driver.find_element(By.NAME, "password")
# Enter email and password
email_input.send_keys(email)
password_input.send_keys(password)

# Submit the login form
print("Logging In")
password_input.send_keys(Keys.ENTER)

# Wait for the login to complete
time.sleep(3)

print("Switching Profiles")
# Navigate to Viewing Activity page
driver.get(f"https://www.netflix.com/SwitchProfile?tkn={profile}")

# Wait for the login to complete
time.sleep(3)

print("Getting Viewing Activity")
# Navigate to Viewing Activity page
driver.get("https://www.netflix.com/viewingactivity")

time.sleep(3)

print("Clicking Download all")
# Navigate to a page and download a file
element = driver.find_element(By.LINK_TEXT, "Download all").click()

print("Sleeping just in case")
time.sleep(10)

driver.quit()



import uuid
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from syftbox.server.settings import ServerSettings
from syftbox.server.sync.file_store import FileStore
from syftbox.server.sync.hash import hash_file


def test_put_atomic(tmpdir):
    settings = ServerSettings.from_data_folder(tmpdir)
    syft_path = Path("test.txt")
    system_path = settings.snapshot_folder / syft_path

    with ThreadPoolExecutor(max_workers=5) as executor:
        executor.map(lambda _: FileStore(settings).put(syft_path, uuid.uuid4().bytes), range(25))

    assert system_path.exists()
    metadata = FileStore(settings).get_metadata(syft_path)
    assert metadata.hash_bytes == hash_file(system_path).hash_bytes



import os
import platform
import subprocess
from typing import Tuple


def open_dir(folder_path) -> Tuple[bool, str]:
    """Open the folder specified by `folder_path` in the default file explorer."""
    if not os.path.exists(folder_path):
        return False, f"Folder does not exist: {folder_path}"

    try:
        if platform.system() == "Darwin":
            subprocess.run(["open", folder_path])
        elif platform.system() == "Windows":
            subprocess.run(["explorer", folder_path])
        elif platform.system() == "Linux":
            # xdg-open is not available on all Linux distros
            distro_explorer = _get_linux_file_explorer()
            subprocess.run([distro_explorer, folder_path])
        else:
            return False, f"Unsupported OS for opening folders: {platform.system()}"
        return True
    except Exception as e:
        return False, str(e)


def _get_linux_file_explorer() -> str:
    """Get the default file explorer for Linux distros."""
    # implement as needed, for now just return xdg-open
    return "xdg-open"



from datetime import timedelta
from pathlib import Path
from typing import Optional

from fastapi import Request
from pydantic import Field, SecretStr, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self, Union

DEV_JWT_SECRET = "changethis"


class ServerSettings(BaseSettings):
    """
    Reads the server settings from the environment variables, using the prefix SYFTBOX_.

    example:
    `export SYFTBOX_DATA_FOLDER=data/data_folder`
    will set the server_settings.data_folder to `data/data_folder`

    see: https://docs.pydantic.dev/latest/concepts/pydantic_settings/#parsing-environment-variable-values
    """

    model_config = SettingsConfigDict(env_prefix="SYFTBOX_", env_file="server.env")
    sendgrid_secret: Optional[SecretStr] = None

    data_folder: Path = Field(default=Path("data").resolve())
    """Absolute path to the server data folder"""

    email_service_api_key: str = Field(default="")
    """API key for the email service"""

    jwt_secret: SecretStr = DEV_JWT_SECRET
    """Secret key for the JWT tokens. Dev secret is not allowed in production"""

    jwt_email_token_exp: timedelta = timedelta(hours=1)
    jwt_access_token_exp: Optional[timedelta] = None
    jwt_algorithm: str = "HS256"
    auth_enabled: bool = False

    @field_validator("data_folder", mode="after")
    def data_folder_abs(cls, v):
        return Path(v).expanduser().resolve()

    @model_validator(mode="after")
    def auth_secret_not_empty(self):
        secret_val = self.jwt_secret.get_secret_value()
        secret_val_is_set = len(secret_val) > 0 and secret_val != DEV_JWT_SECRET

        if self.auth_enabled and not secret_val_is_set:
            raise ValueError("auth is enabled, but jwt_secret is not set")

        # ensure auth is always enabled when jwt_secret is set
        if not self.auth_enabled and secret_val_is_set:
            raise ValueError("jwt_secret is defined, but no_auth is enabled")

        return self

    @model_validator(mode="after")
    def sendgrid_secret_not_empty(self):
        if self.auth_enabled and self.sendgrid_secret is None:
            raise ValueError("auth is enabled, but no sendgrid_secret is defined")

        return self

    @property
    def folders(self) -> list[Path]:
        return [self.data_folder, self.snapshot_folder]

    @property
    def snapshot_folder(self) -> Path:
        return self.data_folder / "snapshot"

    @property
    def logs_folder(self) -> Path:
        return self.data_folder / "logs"

    @property
    def user_file_path(self) -> Path:
        return self.data_folder / "users.json"

    @classmethod
    def from_data_folder(cls, data_folder: Union[Path, str]) -> Self:
        data_folder = Path(data_folder)
        return cls(
            data_folder=data_folder,
        )

    @property
    def file_db_path(self) -> Path:
        return self.data_folder / "file.db"

    def read(self, path: Path) -> bytes:
        with open(self.snapshot_folder / path, "rb") as f:
            return f.read()


def get_server_settings(request: Request) -> ServerSettings:
    return request.state.server_settings



import sys
from datetime import datetime
from pathlib import Path
from shutil import make_archive
from typing import Union

from loguru import logger

from syftbox.lib.constants import DEFAULT_LOGS_DIR
from syftbox.lib.types import PathLike, to_path


def setup_logger(
    level: Union[str, int] = "DEBUG",
    log_dir: PathLike = DEFAULT_LOGS_DIR,
    keep_logs: int = 10,
):
    logger.remove()
    logger.add(level=level, sink=sys.stderr, diagnose=False, backtrace=False)

    # new file per run - no rotation needed
    # always log debug level
    log_file = Path(log_dir, f"syftbox_{int(datetime.now().timestamp())}.log")
    logger.add(
        log_file,
        level="DEBUG",
        rotation=None,
        compression=None,
    )

    # keep last 5 logs
    logs_to_delete = sorted(log_dir.glob("syftbox_*.log"))[:-keep_logs]
    for log in logs_to_delete:
        try:
            log.unlink()
        except Exception:
            pass


def zip_logs(output_path, log_dir: PathLike = DEFAULT_LOGS_DIR):
    logs_folder = to_path(log_dir)
    return make_archive(output_path, "zip", logs_folder)



class FatalSyncError(Exception):
    """Base exception to signal sync should be interrupted."""

    pass


class SyncEnvironmentError(FatalSyncError):
    """the sync environment is corrupted (e.g. sync folder deleted), syncing cannot continue."""

    pass



"""
SyftBox Client Shim for apps and external dependencies


NOTE: this will likely get refactored as it's own SDK.
But we need it to maintain compatibility with apps
"""

from pathlib import Path

from typing_extensions import Optional, Self

from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.types import PathLike, to_path
from syftbox.lib.workspace import SyftWorkspace

# this just makes it a bit clear what the default is for the api_data() method
CURRENT_API_REQUEST_NAME = None
MY_DATASITE = None


class Client:
    """
    Client shim for SyftBox Apps

    Minimal set of properties and methods exposed to the apps.
    """

    def __init__(self, conf: SyftClientConfig):
        self.config = conf
        self.workspace = SyftWorkspace(self.config.data_dir)

    @property
    def email(self):
        """Email of the current user"""
        return self.config.email

    @property
    def config_path(self) -> Path:
        """Path to the config of the current user"""
        return self.config.path

    @property
    def my_datasite(self) -> Path:
        """Path to the datasite of the current user"""
        return self.workspace.datasites / self.config.email

    @property
    def datasites(self) -> Path:
        """Path to the datasites folder"""
        return self.workspace.datasites

    @property
    def sync_folder(self) -> Path:
        """Deprecated property use `client.datasites` instead"""
        return self.workspace.datasites

    @property
    def datasite_path(self) -> Path:
        """Deprecated property. Use `client.my_datasite` instead"""
        return self.workspace.datasites / self.config.email

    @classmethod
    def load(cls, filepath: Optional[PathLike] = None) -> Self:
        """
        Load the client configuration from the given file path or env var or default location
        Raises: ClientConfigException
        """
        return cls(conf=SyftClientConfig.load(filepath))

    @property
    def api_request_name(self) -> str:
        """Returns the name of root directory of the API request calling this property.

        Use this property instead of hardcoding your API request's directory name,
        as SyftBox may dynamically change it to prevent conflicts.
        """
        # The below works coz we set the cwd to the app's path before executing run.sh (see find_and_run_script method)
        api_path = Path.cwd()
        api_name = api_path.name
        return api_name

    def api_data(self, api_request_name: str = CURRENT_API_REQUEST_NAME, datasite: Optional[str] = MY_DATASITE) -> Path:
        """
        Gets the filesystem path to an application's API data directory for a specific datasite.

        Args:
            api_request_name (Optional[str], default=CURRENT_API_REQUEST_NAME): The name of the API request
            whose API data path is needed.
                If None, defaults to the name of the API request from which this method is being called.
            datasite (Optional[str], default=MY_DATASITE): The datasite's email.
                If None, defaults to the current user's configured email.

        Returns:
            Path: A filesystem path pointing to '<workspace>/datasites/<datasite>/api_data/<api_request_name>'.
        """
        api_request_name = api_request_name or self.api_request_name
        datasite = datasite or self.config.email
        return self.workspace.datasites / datasite / "api_data" / api_request_name

    def makedirs(self, *paths: PathLike) -> None:
        """Create directories"""

        for path in paths:
            to_path(path).mkdir(parents=True, exist_ok=True)



from pathlib import Path

import pytest

from syftbox.lib.workspace import SyftWorkspace


@pytest.fixture
def temp_workspace(tmp_path):
    """Fixture to create a temporary workspace directory."""
    return SyftWorkspace(tmp_path / "workspace")


def test_workspace_init():
    # Test with string path
    workspace = SyftWorkspace("/tmp/test")
    assert isinstance(workspace.data_dir, Path)

    # Test with Path object
    path_obj = Path("~/test2")
    workspace = SyftWorkspace(path_obj)

    assert isinstance(workspace.data_dir, Path)
    assert workspace.data_dir.is_absolute()
    assert "~" not in str(workspace.data_dir)


def test_workspace_directory_structure(tmp_path):
    workspace = SyftWorkspace(tmp_path)

    assert not workspace.datasites.exists()
    assert not workspace.plugins.exists()
    assert not workspace.apps.exists()

    # Create directories
    workspace.mkdirs()

    # Verify directory structure
    assert workspace.data_dir.is_dir()
    assert workspace.datasites.is_dir()
    assert workspace.plugins.is_dir()
    assert workspace.apps.is_dir()



from fastapi import Depends, Request
from typing_extensions import Annotated

from syftbox.client.base import SyftClientInterface

__all__ = ["APIContext"]


# Create a dependency for typed access to the client
async def get_context(request: Request) -> SyftClientInterface:
    return request.app.state.client


APIContext = Annotated[SyftClientInterface, Depends(get_context)]



import pytest

from syftbox.client.client2 import SyftClient, run_migration
from syftbox.client.exceptions import SyftBoxAlreadyRunning
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.exceptions import SyftBoxException


def test_client_single_instance(tmp_path):
    """Test that only one client instance can run"""
    config = SyftClientConfig(
        path=tmp_path / "config.json",
        data_dir=tmp_path,
        email="test@example.com",
        server_url="http://localhost:5001",
        client_url="http://localhost:8080",
    )
    client1 = SyftClient(config)
    client2 = SyftClient(config)

    client1.pid.create()

    # check should raise
    with pytest.raises(SyftBoxAlreadyRunning):
        client2.check_pidfile()

    # start should raise
    with pytest.raises(SyftBoxAlreadyRunning):
        client2.start()

    client1.shutdown()


def test_client_init_datasite(mock_config):
    client = SyftClient(mock_config)
    client.init_datasite()

    assert client.datasite.is_dir()
    assert client.public_dir.is_dir()


def test_register_user(mock_config, httpx_mock):
    httpx_mock.add_response(json={"token": "dummy-token"})
    client = SyftClient(mock_config)
    client.register_self()
    assert client.config.token == "dummy-token"


def test_register_user_error(mock_config, httpx_mock):
    httpx_mock.add_response(status_code=503)
    client = SyftClient(mock_config)
    with pytest.raises(SyftBoxException):
        client.register_self()

    assert client.config.token is None


def test_client_paths(tmp_path):
    """Test that client paths are correctly set up"""
    config = SyftClientConfig(
        path=tmp_path / "config.json",
        data_dir=tmp_path,
        email="test@example.com",
        server_url="http://localhost:5001",
        client_url="http://localhost:8080",
    )
    client = SyftClient(config)

    # data_dir should be the root of the client workspace
    assert client.workspace.data_dir == tmp_path


def test_migration(mock_config):
    # setup old datasites
    datasites = ["test@openmined.org", "test2@openmined.org"]
    old_dirs = ["apps", ".syft"] + datasites
    for dir in old_dirs:
        (mock_config.data_dir / dir).mkdir(parents=True)
    (mock_config.data_dir / ".syft" / "local_syncstate.json").touch()

    run_migration(mock_config)

    # check new workspace
    assert (mock_config.data_dir / "apis").is_dir()
    assert (mock_config.data_dir / "plugins").is_dir()
    assert (mock_config.data_dir / "datasites").is_dir()

    # check migrated datasites
    for ds in datasites:
        assert not (mock_config.data_dir / ds).exists()
        assert (mock_config.data_dir / "datasites" / ds).is_dir()

    # check syncstate migration
    assert not (mock_config.data_dir / ".syft").exists()
    assert (mock_config.data_dir / "plugins" / "local_syncstate.json").is_file()



import os
from pathlib import Path
from typing import Iterable, Union

from typing_extensions import TypeAlias

__all__ = ["PathLike", "UserLike", "to_path"]

PathLike: TypeAlias = Union[str, os.PathLike, Path]
UserLike: TypeAlias = Union[str, Iterable[str]]


def to_path(path: PathLike) -> Path:
    return Path(path).expanduser().resolve()



import asyncio
import json
import platform
import shutil
from functools import lru_cache
from pathlib import Path

import httpx
import uvicorn
from loguru import logger
from pid import PidFile, PidFileAlreadyLockedError, PidFileAlreadyRunningError

from syftbox.__version__ import __version__
from syftbox.client.api import create_api
from syftbox.client.base import SyftClientInterface
from syftbox.client.env import syftbox_env
from syftbox.client.exceptions import SyftBoxAlreadyRunning, SyftInitializationError, SyftServerError
from syftbox.client.logger import setup_logger
from syftbox.client.plugins.apps import AppRunner
from syftbox.client.plugins.sync.manager import SyncManager
from syftbox.client.utils import error_reporting, file_manager, macos
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.datasite import create_datasite
from syftbox.lib.exceptions import SyftBoxException
from syftbox.lib.ignore import IGNORE_FILENAME
from syftbox.lib.workspace import SyftWorkspace

SCRIPT_DIR = Path(__file__).parent
ASSETS_FOLDER = SCRIPT_DIR.parent / "assets"
ICON_FOLDER = ASSETS_FOLDER / "icon"
METADATA_FILENAME = ".metadata.json"


class SyftClient:
    """The SyftBox Client

    This is the main SyftBox client that handles workspace data, server
    communication, and local API services. Only one client instance can run
    for a given workspace directory.

    Warning:
        This class should not be imported directly by sub-systems.
        Use the provided interfaces and context objects instead.

    Raises:
        SyftBoxAlreadyRunning: If another client is already running for the same workspace
        Exception: If the client fails to start due to any reason
    """

    def __init__(self, config: SyftClientConfig, log_level: str = "INFO", **kwargs):
        self.config = config
        self.log_level = log_level

        self.workspace = SyftWorkspace(self.config.data_dir)
        self.pid = PidFile(pidname="syftbox.pid", piddir=self.workspace.data_dir)

        self.server_client = httpx.Client(
            base_url=str(self.config.server_url),
            follow_redirects=True,
            headers=self._server_headers,
        )

        # kwargs for making customization/unit testing easier
        # this will be replaced with a sophisticated plugin system
        self.__sync_manager: SyncManager = kwargs.get("sync_manager", None)
        self.__app_runner: AppRunner = kwargs.get("app_runner", None)
        self.__local_server: uvicorn.Server = None

    @property
    def _server_headers(self):
        # TODO make access token required for initializing the client
        headers = {"email": self.config.email}
        if self.config.access_token is not None:
            headers["Authorization"] = f"Bearer {self.config.access_token}"
        return headers

    @property
    def sync_manager(self):
        """the sync manager. lazily initialized"""
        if self.__sync_manager is None:
            try:
                self.__sync_manager = SyncManager(self.as_context())
            except Exception as e:
                raise SyftInitializationError(f"Failed to initialize sync manager - {e}") from e
        return self.__sync_manager

    @property
    def app_runner(self):
        """the app runner. lazily initialized"""
        if self.__app_runner is None:
            try:
                self.__app_runner = AppRunner(self.as_context())
            except Exception as e:
                raise SyftInitializationError(f"Failed to initialize app runner - {e}") from e
        return self.__app_runner

    @property
    def is_registered(self) -> bool:
        """Check if the current user is registered with the server"""
        return bool(self.config.token)

    @property
    def datasite(self) -> Path:
        """The datasite of the current user"""
        return self.workspace.datasites / self.config.email

    @property
    def public_dir(self) -> Path:
        """The public directory in the datasite of the current user"""
        return self.datasite / "public"

    def start(self):
        try:
            self.pid.create()
        except PidFileAlreadyLockedError:
            raise SyftBoxAlreadyRunning(f"Another instance of SyftBox is running on {self.config.data_dir}")
        self.create_metadata_file()

        logger.info("Started SyftBox client")

        self.config.save()  # commit config changes (like migration) to disk after PID is created
        self.workspace.mkdirs()  # create the workspace directories
        self.register_self()  # register the email with the server
        self.init_datasite()  # init the datasite on local machine

        # start plugins/components
        self.sync_manager.start()
        self.app_runner.start()
        return self.__run_local_server()

    @property
    def metadata_path(self) -> Path:
        return self.workspace.data_dir / METADATA_FILENAME

    def create_metadata_file(self) -> None:
        metadata_json = self.config.model_dump(mode="json")
        metadata_json["version"] = __version__
        self.metadata_path.write_text(json.dumps(metadata_json, indent=2))

    def shutdown(self):
        if self.__local_server:
            _result = asyncio.run(self.__local_server.shutdown())

        if self.__sync_manager:
            self.__sync_manager.stop()

        if self.__app_runner:
            self.__app_runner.stop()

        self.pid.close()
        logger.info("SyftBox client shutdown complete")

    def check_pidfile(self) -> str:
        """Check if another instance of SyftBox is running"""

        try:
            return self.pid.check()
        except PidFileAlreadyRunningError:
            raise SyftBoxAlreadyRunning(f"Another instance of SyftBox is running on {self.config.data_dir}")

    def init_datasite(self):
        if self.datasite.exists():
            return
        create_datasite(self.workspace.datasites, self.config.email)

    def register_self(self):
        """Register the user's email with the SyftBox cache server"""
        if self.is_registered:
            return
        try:
            token = self.__register_email()
            # TODO + FIXME - once we have JWT, we should not store token in config!
            # ideally in OS keychain (using keyring) or
            # in a separate location under self.workspace.plugins
            self.config.token = str(token)
            self.config.save()
            logger.info("Email registration successful")
        except Exception as e:
            raise SyftBoxException(f"Failed to register with the server - {e}") from e

    @lru_cache(1)
    def as_context(self) -> "SyftClientContext":
        """Return a implementation of SyftClientInterface to be injected into sub-systems"""
        return SyftClientContext(self.config, self.workspace, self.server_client)

    def __run_local_server(self):
        logger.info(f"Starting local server on {self.config.client_url}")
        app = create_api(self.as_context())
        self.__local_server = uvicorn.Server(
            config=uvicorn.Config(
                app=app,
                host=self.config.client_url.host,
                port=self.config.client_url.port,
                log_level=self.log_level.lower(),
            )
        )
        return self.__local_server.run()

    def __register_email(self) -> str:
        # TODO - this should probably be wrapped in a SyftCacheServer API?
        response = self.server_client.post("/register", json={"email": self.config.email})
        response.raise_for_status()
        return response.json().get("token")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    # utils
    def open_datasites_dir(self):
        file_manager.open_dir(self.workspace.datasites)

    def copy_icons(self):
        self.workspace.mkdirs()
        if platform.system() == "Darwin":
            macos.copy_icon_file(ICON_FOLDER, self.workspace.data_dir)


class SyftClientContext(SyftClientInterface):
    """
    Concrete implementation of SyftClientInterface that provides a lightweight
    client context for components.

    This class encapsulates the minimal set of attributes needed by components
    without exposing the full SyftClient implementation.

    It will be instantiated by SyftClient, but sub-systems can freely pass it around.
    """

    def __init__(
        self,
        config: SyftClientConfig,
        workspace: SyftWorkspace,
        server_client: httpx.Client,
    ):
        self.config = config
        self.workspace = workspace
        self.server_client = server_client

    @property
    def email(self) -> str:
        return self.config.email

    @property
    def datasite(self) -> Path:
        return self.workspace.datasites / self.config.email

    @property
    def all_datasites(self) -> list[str]:
        """List all datasites in the workspace"""
        return [d.name for d in self.workspace.datasites.iterdir() if (d.is_dir() and "@" in d.name)]

    def __repr__(self) -> str:
        return f"SyftClientContext<{self.config.email}, {self.config.data_dir}>"

    def log_analytics_event(self, event_name: str, **kwargs) -> None:
        """Log an event to the server"""
        event_data = {
            "event_name": event_name,
            **kwargs,
        }

        response = self.server_client.post("/log_event", json=event_data)
        if response.status_code != 200:
            raise SyftServerError(f"Failed to log event: {response.text}")


def run_apps_to_api_migration(new_ws: SyftWorkspace):
    old_sync_folder = new_ws.data_dir
    old_apps_dir = old_sync_folder / "apps"
    new_apps_dir = new_ws.apps

    if old_apps_dir.exists():
        logger.info(f"Migrating directory apps —> {new_apps_dir.relative_to(new_ws.data_dir)}...")
        if new_apps_dir.exists():
            shutil.rmtree(new_apps_dir)
        shutil.move(str(old_apps_dir), str(new_apps_dir))


def run_migration(config: SyftClientConfig, migrate_datasite=True):
    # first run config migration
    config.migrate()

    # then run workspace migration
    new_ws = SyftWorkspace(config.data_dir)

    # migrate workspace/apps to workspace/apis
    run_apps_to_api_migration(new_ws)

    # check for old dir structure and migrate to new
    # data_dir == sync_folder
    old_sync_folder = new_ws.data_dir
    old_datasite_path = Path(old_sync_folder, config.email)

    if not migrate_datasite:
        return

    # Option 2: if syftbox folder has old structure, migrate to new
    if old_datasite_path.exists():
        logger.info("Migrating to new datasite structure")
        new_ws.mkdirs()

        # create the datasites directory & move all under it
        for dir in old_sync_folder.glob("*@*"):
            shutil.move(str(dir), str(new_ws.datasites))

        # move syftignore file
        old_ignore_file = old_sync_folder / IGNORE_FILENAME
        if old_ignore_file.exists():
            shutil.move(str(old_ignore_file), str(new_ws.datasites / IGNORE_FILENAME))

        # move old sync state file
        old_sync_state = old_sync_folder / ".syft" / "local_syncstate.json"
        if old_sync_state.exists():
            shutil.move(str(old_sync_state), str(new_ws.plugins / "local_syncstate.json"))
        if old_sync_state.parent.exists():
            shutil.rmtree(str(old_sync_state.parent))


def run_client(
    client_config: SyftClientConfig, open_dir: bool = False, log_level: str = "INFO", migrate_datasite=True
) -> int:
    """Run the SyftBox client"""
    client = None

    setup_logger(log_level, log_dir=client_config.data_dir / "logs")

    error_config = error_reporting.make_error_report(client_config)
    logger.info(f"Client metadata\n{error_config.model_dump_json(indent=2)}")

    # a flag to disable icons
    # GitHub CI needs to zip sync dir in tests and fails when it encounters Icon\r files
    if syftbox_env.DISABLE_ICONS:
        logger.debug("Directory icons are disabled")

    try:
        client = SyftClient(client_config, log_level=log_level)
        # we don't want to run migration if another instance of client is already running
        bool(client.check_pidfile()) and run_migration(client_config, migrate_datasite=migrate_datasite)
        (not syftbox_env.DISABLE_ICONS) and client.copy_icons()
        open_dir and client.open_datasites_dir()
        client.start()
    except SyftBoxAlreadyRunning as e:
        logger.error(e)
        return -1
    except KeyboardInterrupt:
        logger.info("Received keyboard interrupt. Shutting down the client")
    except Exception as e:
        logger.exception("Unhandled exception when starting the client", e)
        return -2
    finally:
        client and client.shutdown()
    return 0



from pathlib import Path

from typing_extensions import Callable, TypeAlias
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

__all__ = [
    "FSWatchdog",
    "AnyFileSystemEventHandler",
    "FSEventCallbacks",
    "FileSystemEvent",
]

FSEventCallbacks: TypeAlias = list[Callable[[FileSystemEvent], None]]


class FSWatchdog:
    def __init__(self, watch_dir: Path, event_handler: FileSystemEventHandler):
        self.watch_dir = watch_dir
        self.event_handler = event_handler
        self._observer = Observer()

    def start(self):
        # observer starts it's own thread
        self._observer.schedule(
            self.event_handler,
            self.watch_dir,
            recursive=True,
        )
        self._observer.start()

    def stop(self):
        self._observer.stop()
        self._observer.join()


class AnyFileSystemEventHandler(FileSystemEventHandler):
    def __init__(
        self,
        watch_dir: Path,
        callbacks: FSEventCallbacks,
        ignored: list[str] = [],
    ):
        self.watch_dir = watch_dir
        self.callbacks = callbacks
        self.ignored = [Path(self.watch_dir, ignore) for ignore in ignored]

    def on_any_event(self, event: FileSystemEvent) -> None:
        for ignore in self.ignored:
            if event.src_path.startswith(str(ignore)):
                return

        for cb in self.callbacks:
            cb(event)



from pathlib import Path
from typing import Any, Generator

import pytest

from syftbox.lib.client_shim import SyftClientConfig


@pytest.fixture
def mock_config(monkeypatch, tmp_path) -> Generator[SyftClientConfig, Any, None]:
    config_path = Path(tmp_path, "config.json")
    data_dir = Path(tmp_path)
    conf = SyftClientConfig(
        path=config_path,
        data_dir=data_dir,
        email="test@openmined.org",
        client_url="http://test:8080",
    )
    conf.save()
    conf.data_dir.mkdir(parents=True, exist_ok=True)

    def mock_load(*args, **kwargs):
        nonlocal conf
        return conf

    monkeypatch.setattr(SyftClientConfig, "load", mock_load)

    yield conf

    monkeypatch.undo()



from syftbox.lib.debug import debug_report, debug_report_yaml


def test_debug_report(mock_config):
    result = debug_report()
    assert isinstance(result, dict)
    assert "system" in result
    assert "syftbox" in result
    assert "syftbox_env" in result
    assert "resources" in result["system"]
    assert "operating_system" in result["system"]
    assert "python" in result["system"]
    assert "command" in result["syftbox"]
    assert "client_config_path" in result["syftbox"]
    assert "client_config" in result["syftbox"]
    assert "apps_dir" in result["syftbox"]
    assert "apps" in result["syftbox"]


def test_debug_report_readable(mock_config):
    result = debug_report_yaml()
    assert isinstance(result, str)
    assert "system" in result
    assert "syftbox" in result
    assert "syftbox_env" in result
    assert "resources" in result
    assert "operating_system" in result
    assert "python" in result
    assert "command" in result
    assert "client_config_path" in result
    assert "client_config" in result
    assert "apps_dir" in result
    assert "apps" in result



# Example test using the framework

import asyncio
import json
import secrets
import shutil
from pathlib import Path

import pytest
from loguru import logger

from tests.e2e.conftest import Client, E2EContext, Server

AGGREGATOR_CONFIG = {
    "participants": ["user1@openmined.org", "user2@openmined.org", "user3@openmined.org"],
}
MAX_COPY_MODELS = 2


def deployment_config():
    return {
        "e2e_name": "model_aggregator",
        "server": Server(port=5001),
        "clients": [
            Client(
                name="agg",
                port=8080,
                apps=["https://github.com/OpenMined/pretrained_model_aggregator"],
            ),
            Client(
                name="user1",
                port=8081,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/pretrained_model_local",
                ],
            ),
            Client(
                name="user2",
                port=8082,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/pretrained_model_local",
                ],
            ),
            Client(
                name="user3",
                port=8083,
                server_port=5001,
                apps=[
                    "https://github.com/OpenMined/pretrained_model_local",
                ],
            ),
        ],
    }


async def copy_model_to_public(e2e_context: E2EContext, client: Client):
    await e2e_context.wait_for_api("pretrained_model_local", client)
    pretrained_models_dir = client.api_path("pretrained_model_local") / "pretrained_models"
    all_models = list(pretrained_models_dir.glob("*.pt"))
    for _ in range(MAX_COPY_MODELS):
        # random.random is not concurrency safe, using secrets.randbelow
        rand_idx = secrets.randbelow(len(all_models))
        random_model = all_models[rand_idx]
        logger.debug(f"Copying {random_model} to {client.email} public dir")
        shutil.copy(random_model, client.public_dir)
        assert Path(client.public_dir, random_model.name).exists()


@pytest.mark.asyncio
@pytest.mark.parametrize("e2e_context", [deployment_config()], indirect=True, ids=["model_aggregator"])
async def test_e2e_model_aggregator(e2e_context: E2EContext):
    logger.info(f"Starting E2E '{e2e_context.e2e_name}'")
    e2e_context.reset_test_dir()
    agg_client = e2e_context.clients[0]

    await e2e_context.start_all()
    await e2e_context.wait_for_api("pretrained_model_aggregator", agg_client)

    # copy launch config
    logger.info("Copying launch config")
    launch_dir = agg_client.api_data_dir("pretrained_model_aggregator") / "launch"
    launch_dir.mkdir(parents=True, exist_ok=True)
    participants_file = launch_dir / "participants.json"
    participants_file.write_text(json.dumps(AGGREGATOR_CONFIG))

    # copy model to public dir
    logger.info("Copying models for all users in their public directories")
    await asyncio.gather(*[copy_model_to_public(e2e_context, client) for client in e2e_context.clients[1:]])

    # wait for results
    logger.info("Waiting for aggregator to generate results")
    done_dir = agg_client.api_data_dir("pretrained_model_aggregator") / "done"
    results_file = done_dir / "results.json"
    await e2e_context.wait_for_path(results_file, timeout=120, interval=1)

    # check results
    result = json.loads(results_file.read_text())
    logger.info(f"Validating results\n{result}")
    assert result["accuracy"] >= 10.0
    assert set(result["participants"]) == set(AGGREGATOR_CONFIG["participants"])
    assert len(result["missing_peers"]) == 0



import asyncio
import json
import random

import pytest
from loguru import logger

from tests.e2e.conftest import Client, E2EContext, Server

INIT_DATA = {
    "participants": [
        "user1@openmined.org",
        "user2@openmined.org",
        "user1@openmined.org",
    ],
    "data": 0,
    "current_index": 0,
}


def deployment_config():
    return {
        "e2e_name": "ring",
        "server": Server(port=5001),
        "clients": [
            Client(name="user1", port=8080, server_port=5001, apps=["https://github.com/OpenMined/ring"]),
            Client(name="user2", port=8081, server_port=5001, apps=["https://github.com/OpenMined/ring"]),
        ],
    }


@pytest.mark.asyncio
@pytest.mark.parametrize("e2e_context", [deployment_config()], indirect=True, ids=["deployment"])
async def test_e2e_ring(e2e_context: E2EContext):
    logger.info(f"Starting E2E '{e2e_context.e2e_name}'")
    e2e_context.reset_test_dir()
    await e2e_context.start_all()

    # wait for all clients to install ring
    logger.info("Waiting for clients to install ring app")
    await asyncio.gather(*[e2e_context.wait_for_api("ring", client) for client in e2e_context.clients])

    values = []

    # copy secrets to all clients
    logger.info("Copying secrets to all clients")
    for client in e2e_context.clients:
        val = random.randint(0, 100)
        api_path = client.api_path("ring")
        api_path.joinpath("secret.json").write_text(json.dumps({"data": val}))
        values.append(val)
        logger.debug(f"{client.name} secret: {val}")

    # kickstart ring
    logger.info("Initiating ring by copying data.json")
    init_user = e2e_context.clients[0]
    running_dir = init_user.api_data_dir("ring") / "running"
    running_dir.mkdir(parents=True, exist_ok=True)
    running_dir.joinpath("data.json").write_text(json.dumps(INIT_DATA))

    logger.info("Waiting for ring results to be available")
    output = init_user.api_data_dir("ring") / "done" / "data.json"
    await e2e_context.wait_for_path(output, timeout=120, interval=1)

    # check the output
    result = json.loads(output.read_text())
    logger.info(f"Validating results\n{result}")
    assert result["data"] == sum(values) + values[0]
    assert result["current_index"] == 2



import pytest
from typer.testing import CliRunner

from syftbox.app.cli import app as app_cli
from syftbox.app.install import InstallResult

# Initialize test runner
runner = CliRunner()


@pytest.fixture
def mock_apps_dir(mock_config):
    apps_dir = mock_config.data_dir / "apis"
    apps_dir.mkdir(exist_ok=True)
    yield apps_dir


@pytest.fixture
def mock_app_list(mock_apps_dir):
    apps = [mock_apps_dir / "app1", mock_apps_dir / "app2"]
    for app in apps:
        app.mkdir(exist_ok=True)
    yield (mock_apps_dir, apps)


@pytest.fixture
def mocked_app_install(monkeypatch, mock_apps_dir):
    def mock_install(config, repository, branch):
        return InstallResult(
            app_name=repository,
            app_path=mock_apps_dir / repository,
            error=None,
            details=None,
        )

    monkeypatch.setattr("syftbox.app.cli.install_app", mock_install)
    yield
    monkeypatch.undo()


@pytest.fixture
def mocked_app_uninstall(monkeypatch, mock_config):
    def mock_uninstall(*args, **kwargs):
        return mock_config.data_dir / "apps"

    monkeypatch.setattr("syftbox.app.cli.uninstall_app", mock_uninstall)
    yield
    monkeypatch.undo()


def test_list_app(mock_app_list):
    result = runner.invoke(app_cli, ["list"])
    assert result.exit_code == 0
    assert "Apps installed" in result.stdout
    assert "app1" in result.stdout
    assert "app2" in result.stdout


def test_list_app_empty(mock_apps_dir):
    result = runner.invoke(app_cli, ["list"])
    assert result.exit_code == 0
    assert "No apps installed" in result.stdout


def test_install_app(mocked_app_install):
    result = runner.invoke(app_cli, ["install", "OpenMined/tutorials-app"])
    assert result.exit_code == 0
    assert "tutorials-app" in result.stdout


def test_uninstall_app(mocked_app_uninstall):
    result = runner.invoke(app_cli, ["uninstall", "app1"])
    assert result.exit_code == 0
    assert "Uninstalled app" in result.stdout
    assert "app1" in result.stdout



import json

import pytest

from syftbox.app.install import load_config


def test_load_app_config(tmp_path):
    valid_json_config = {
        "version": "0.1.0",
        "app": {
            "version": "0.1.0",
            "run": {"command": ["python", "main.py"], "interval": "10"},
            "env": {},
            "platforms": ["linux"],
            "pre_install": ["pip", "install", "psutil"],
            "post_install": [],
            "pre_update": [],
            "post_update": [],
        },
    }

    app_conf = tmp_path / "app.json"
    app_conf.write_text(json.dumps(valid_json_config))

    app_config = load_config(app_conf)
    assert app_config.version == valid_json_config["version"]
    assert app_config.app.version == valid_json_config["app"]["version"]
    assert app_config.app.run.command == valid_json_config["app"]["run"]["command"]
    assert vars(app_config.app.env) == valid_json_config["app"]["env"]
    assert app_config.app.platforms == valid_json_config["app"]["platforms"]
    assert app_config.app.pre_install == valid_json_config["app"]["pre_install"]
    assert app_config.app.pre_update == valid_json_config["app"]["pre_update"]
    assert app_config.app.post_update == valid_json_config["app"]["post_update"]


def test_load_invalid_app_config(tmp_path):
    app_conf = tmp_path / "app.json"
    app_conf.write_text("Invalid JSON")

    with pytest.raises(ValueError) as expt:
        load_config(app_conf)
        assert expt.value == "File isn't in JSON format"


def test_load_inexistent_app_config():
    with pytest.raises(ValueError) as expt:
        load_config("inexistent_app.json")
        assert expt.value == "Couln't find the json config file for this path."



from pathlib import Path

from rich import print as rprint
from typer import Context, Exit, Option, Typer
from typing_extensions import Annotated

from syftbox.lib.constants import DEFAULT_CONFIG_PATH, DEFAULT_DATA_DIR, DEFAULT_PORT, DEFAULT_SERVER_URL

app = Typer(
    name="SyftBox Client",
    pretty_exceptions_enable=False,
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)

# Define options separately to keep the function signature clean
# fmt: off

# client commands opts
CLIENT_PANEL = "Client Options"
LOCAL_SERVER_PANEL = "Local Server Options"

EMAIL_OPTS = Option(
    "-e", "--email",
    rich_help_panel=CLIENT_PANEL,
    help="Email for the SyftBox datasite",
)
SERVER_OPTS = Option(
    "-s", "--server",
    rich_help_panel=CLIENT_PANEL,
    help="SyftBox cache server URL",
)
DATA_DIR_OPTS = Option(
    "-d", "--data-dir", "--sync_folder",
    rich_help_panel=CLIENT_PANEL,
    help="Directory where SyftBox stores data",
)
CONFIG_OPTS = Option(
    "-c", "--config", "--config_path",
    rich_help_panel=CLIENT_PANEL,
    help="Path to SyftBox configuration file",
)
OPEN_OPTS = Option(
    is_flag=True,
    rich_help_panel=CLIENT_PANEL,
    help="Open SyftBox sync/data dir folder on client start",
)
PORT_OPTS = Option(
    "-p", "--port",
    rich_help_panel=LOCAL_SERVER_PANEL,
    help="Local port for the SyftBox client",
)
RELOAD_OPTS = Option(
    rich_help_panel=LOCAL_SERVER_PANEL,
    help="Run server in hot reload. Should not see this in production",
)
VERBOSE_OPTS = Option(
    "-v", "--verbose",
    is_flag=True,
    help="Enable verbose mode",
)



TOKEN_OPTS = Option(
    "--token",
    help="Token for password reset",
)

# report command opts
REPORT_PATH_OPTS = Option(
    "-o", "--output-dir",
    help="Directory to save the log file",
)

# fmt: on


@app.callback(invoke_without_command=True)
def client(
    ctx: Context,
    data_dir: Annotated[Path, DATA_DIR_OPTS] = DEFAULT_DATA_DIR,
    email: Annotated[str, EMAIL_OPTS] = None,
    server: Annotated[str, SERVER_OPTS] = DEFAULT_SERVER_URL,
    config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
    port: Annotated[int, PORT_OPTS] = DEFAULT_PORT,
    open_dir: Annotated[bool, OPEN_OPTS] = True,
    verbose: Annotated[bool, VERBOSE_OPTS] = False,
):
    """Run the SyftBox client"""

    if ctx.invoked_subcommand is not None:
        # If a subcommand is being invoked, just return
        return

    # lazy import to imporve cli startup speed
    from syftbox.client.cli_setup import get_migration_decision, setup_config_interactive
    from syftbox.client.client2 import run_client
    from syftbox.client.utils.net import get_free_port, is_port_in_use

    if port == 0:
        port = get_free_port()
    elif is_port_in_use(port):
        # new_port = get_free_port()
        # port = new_port
        rprint(f"[bold red]Error:[/bold red] Client cannot start because port {port} is already in use!")
        raise Exit(1)

    client_config = setup_config_interactive(config_path, email, data_dir, server, port)

    migrate_datasite = get_migration_decision(client_config.data_dir)

    log_level = "DEBUG" if verbose else "INFO"
    code = run_client(
        client_config=client_config, open_dir=open_dir, log_level=log_level, migrate_datasite=migrate_datasite
    )
    raise Exit(code)


@app.command()
def report(
    output_path: Annotated[Path, REPORT_PATH_OPTS] = Path(".").resolve(),
    config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
):
    """Generate a report of the SyftBox client"""
    from datetime import datetime

    from syftbox.client.logger import zip_logs
    from syftbox.lib.client_config import SyftClientConfig

    try:
        config = SyftClientConfig.load(config_path)
        name = f"syftbox_logs_{datetime.now().strftime('%Y_%m_%d_%H%M')}"
        output_path = Path(output_path, name).resolve()
        output_path_with_extension = zip_logs(output_path, log_dir=config.data_dir / "logs")
        rprint(f"Logs from {config.data_dir} saved at {output_path_with_extension}.")
    except Exception as e:
        rprint(f"[red]Error[/red]: {e}")
        raise Exit(1)


def main():
    app()


if __name__ == "__main__":
    main()



import json
import logging
import sys
from pathlib import Path
from typing import Union

from loguru import logger

custom_format = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level}</level> | <cyan>{message}</cyan>"

ANALYTICS_EVENT = "analytics_event"


def _default_logger_filter(record: dict):
    return record["extra"].get("event_type") != ANALYTICS_EVENT


def _analytics_logger_filter(record: dict):
    return record["extra"].get("event_type") == ANALYTICS_EVENT


analytics_logger = logger.bind(event_type=ANALYTICS_EVENT)


def analytics_formatter(record: dict):
    serialized = json.dumps(record["extra"])
    record["extra"]["serialized"] = serialized
    return "{extra[serialized]}\n"


def setup_logger(logs_folder: Path, level: Union[str, int] = "DEBUG"):
    logs_folder.mkdir(parents=True, exist_ok=True)

    logger.remove()

    # Standard server logs
    logger.add(
        level=level,
        sink=sys.stderr,
        diagnose=False,
        backtrace=False,
        format=custom_format,
        filter=_default_logger_filter,
    )

    logger.add(
        logs_folder / "server.log",
        rotation="100 MB",  # Rotate after the log file reaches 100 MB
        retention=2,  # Keep only the last 1 log files
        compression="zip",  # Usually, 10x reduction in file size
        filter=_default_logger_filter,
    )

    # Dedicated logger for analytics events
    # example usage: user_event_logger.info("User logged in")
    logger.add(
        logs_folder / "analytics.log",
        rotation="100 MB",
        compression="zip",
        format=analytics_formatter,
        filter=_analytics_logger_filter,
    )

    uvicorn_access = logging.getLogger("uvicorn.access")
    uvicorn_access.disabled = True

    logger.info(f"Logger set up. Saving logs to {logs_folder}")



import sys
from pathlib import Path

from rich import print as rprint
from typer import Argument, Exit, Option, Typer
from typing_extensions import Annotated

from syftbox.__version__ import __version__
from syftbox.app.manager import install_app, list_app, uninstall_app
from syftbox.client.base import SyftClientInterface
from syftbox.client.client2 import SyftClient
from syftbox.client.plugins.apps import find_and_run_script
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.constants import DEFAULT_CONFIG_PATH
from syftbox.lib.exceptions import ClientConfigException
from syftbox.lib.workspace import SyftWorkspace

app = Typer(
    name="SyftBox Apps",
    help="Manage SyftBox apps",
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)

CONFIG_OPTS = Option("-c", "--config", "--config_path", help="Path to the SyftBox config")
CALLED_BY_OPTS = Option("--called-by", help="The context from which the command is called", hidden=True)
REPO_ARGS = Argument(..., show_default=False, help="SyftBox App git repo URL")
BRANCH_OPTS = Option("-b", "--branch", help="git branch name")
UNINSTALL_ARGS = Argument(..., show_default=False, help="Name of the SyftBox App to uninstall")
APP_ENV_SCRIPT = """
if [ ! -d .venv ]; then
    uv venv .venv
fi
. .venv/bin/activate
"""


@app.command()
def list(config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH):
    """List all installed Syftbox apps"""
    workspace = get_workspace(config_path)
    result = list_app(workspace)

    if len(result.apps) == 0:
        rprint(f"No apps installed in '{result.apps_dir}'")
        sys.exit(0)

    rprint(f"Apps installed in '{result.apps_dir}'")
    for app in result.apps:
        rprint(f"- [bold cyan]{app.name}[/bold cyan]")


@app.command()
def install(
    repository: Annotated[str, REPO_ARGS],
    branch: Annotated[str, BRANCH_OPTS] = "main",
    config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
    called_by: Annotated[str, CALLED_BY_OPTS] = "user",
):
    """Install a new Syftbox app"""
    client = get_client(config_path)
    result = install_app(client.workspace, repository, branch)
    if result.error:
        rprint(f"[bold red]Error:[/bold red] {result.error}")
        raise Exit(1)

    try:
        client.log_analytics_event("app_install", app_name=result.app_name, called_by=called_by)
    except Exception:
        pass

    rprint(f"Installed app [bold]'{result.app_name}'[/bold]\nLocation: '{result.app_path}'")


@app.command()
def uninstall(
    app_name: Annotated[str, UNINSTALL_ARGS],
    config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
):
    """Uninstall a Syftbox app"""
    workspace = get_workspace(config_path)
    result = uninstall_app(app_name, workspace)
    if not result:
        rprint(f"[bold red]Error:[/bold red] '{app_name}' app not found")
        raise Exit(1)

    rprint(f"Uninstalled app [bold]'{app_name}'[/bold] from '{result}'")


@app.command()
def run(
    app_name: str,
    config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
):
    """Run a Syftbox app"""
    workspace = get_workspace(config_path)

    extra_args = []
    try:
        rprint(f"Running [bold]'{app_name}'[/bold]\nLocation: '{workspace.apps}'\n")
        result = find_and_run_script(str(workspace.apps / app_name), extra_args, str(config_path))
        rprint("[bold yellow]stdout:[/bold yellow]")
        print(result.stdout)
        rprint("[bold yellow]stderr:[/bold yellow]")
        print(result.stderr)
    except Exception as e:
        rprint("[bold red]Error:[/bold red]", e)
        raise Exit(1)


@app.command(rich_help_panel="General Options")
def env(with_syftbox: bool = False):
    """Setup virtual env for app. With option to install syftbox matching client version"""

    script = APP_ENV_SCRIPT
    if with_syftbox:
        script += f"\nuv pip install -U syftbox=={__version__}"
    print(script)


# @app.command()
# def update(
#     app_name: Annotated[str, Argument(help="Name of the app to uninstall")],
#     config_path: Annotated[Path, CONFIG_OPTS] = DEFAULT_CONFIG_PATH,
# ):
#     """Update a Syftbox app"""
#     pass


def get_client(config_path: Path) -> SyftClientInterface:
    try:
        conf = SyftClientConfig.load(config_path)
        return SyftClient(conf).as_context()
    except ClientConfigException:
        msg = (
            f"[bold red]Error:[/bold red] Couldn't load config at: [yellow]'{config_path}'[/yellow]\n"
            "Please ensure that:\n"
            "  - The configuration file exists at the specified path.\n"
            "  - You've run the SyftBox atleast once.\n"
            f"  - For custom configs, provide the proper path using [cyan]--config[/cyan] flag"
        )
        rprint(msg)
        raise Exit(1)
    except Exception as e:
        rprint(f"[bold red]Error:[/bold red] {e}")
        raise Exit(1)


def get_workspace(config_path: Path) -> SyftWorkspace:
    try:
        conf = SyftClientConfig.load(config_path)
        return SyftWorkspace(conf.data_dir)
    except ClientConfigException:
        msg = (
            f"[bold red]Error:[/bold red] Couldn't load config at: [yellow]'{config_path}'[/yellow]\n"
            "Please ensure that:\n"
            "  - The configuration file exists at the specified path.\n"
            "  - You've run the SyftBox atleast once.\n"
            f"  - For custom configs, provide the proper path using [cyan]--config[/cyan] flag"
        )
        rprint(msg)
        raise Exit(1)
    except Exception as e:
        rprint(f"[bold red]Error:[/bold red] {e}")
        raise Exit(1)



from pathlib import Path

from loguru import logger

from syftbox.lib.constants import PERM_FILE
from syftbox.lib.exceptions import SyftBoxException
from syftbox.lib.ignore import create_default_ignore_file
from syftbox.lib.lib import SyftPermission


def create_datasite(datasite_root: Path, email: str):
    # Create workspace/datasites/.syftignore
    create_default_ignore_file(datasite_root)

    user_root = datasite_root / email
    user_public_dir = user_root / "public"

    # Create perm file for the datasite
    if not user_root.is_dir():
        try:
            logger.info(f"creating datasite at {user_root}")
            user_root.mkdir(parents=True, exist_ok=True)
            perms = SyftPermission.datasite_default(email)
            perms.save(str(user_root / PERM_FILE))
        except Exception as e:
            # this is a problematic scenario - probably because you can't setup the basic
            # datasite structure. So, we should probably just exit here.
            raise SyftBoxException(f"Failed to initialize datasite - {e}") from e

    if not user_public_dir.is_dir():
        try:
            logger.info(f"creating public dir in datasite at {user_public_dir}")
            user_public_dir.mkdir(parents=True, exist_ok=True)
            perms = SyftPermission.mine_with_public_read(email)
            perms.save(str(user_public_dir / PERM_FILE))
        except Exception as e:
            # not a big deal if we can't create the public folder
            # more likely that the above step fails than this
            logger.exception("Failed to create folder with public perms", e)



from fastapi import APIRouter
from fastapi.responses import PlainTextResponse

from syftbox.__version__ import __version__
from syftbox.client.routers.common import APIContext

router = APIRouter()


@router.get("/")
async def index():
    return PlainTextResponse(f"SyftBox {__version__}")


@router.get("/version")
async def version():
    return {"version": __version__}


@router.get("/metadata")
async def metadata(ctx: APIContext):
    return {"datasite": ctx.email}



import httpx
from jinja2 import Template
from loguru import logger

from syftbox.server.settings import ServerSettings

SENDER_EMAIL = "SyftBox <auth@openmined.org>"
SENDGRID_SERVER = "https://api.sendgrid.com/v3/mail/send"
SMTP_PORT = 465

token_email_template = """
<!doctype html>
<html>
  <head>
    <style>
      body {
        font-family: Arial, sans-serif;
        background-color: #f4f4f4;
        color: #333;
        margin: 0;
        padding: 0;
      }
      .container {
        max-width: 600px;
        margin: 50px auto;
        background-color: #ffffff;
        padding: 20px;
        border-radius: 8px;
        box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
      }
      h1 {
        color: #333;
        text-align: center;
      }
      p {
        font-size: 16px;
        line-height: 1.5;
      }
      code {
        display: block;
        background-color: #f0f0f0;
        color: #ff8c00;
        padding: 10px;
        font-size: 14px;
        margin: 20px auto;
        border-radius: 4px;
        width: 90%;
        word-wrap: break-word;
      }
      .footer {
        text-align: center;
        font-size: 12px;
        color: #aaa;
        margin-top: 20px;
      }
    </style>
  </head>
  <body>
    <div class="container">
      <h1>Welcome!</h1>
      <p>
        Use the following token in your CLI to complete your registration:
      </p>
      <code> {{ token }} </code>
      <p>If you did not request this, please ignore this email.</p>
    </div>
  </body>
</html>
"""

reset_password_token_email_template = """
<!doctype html>
<html>
  <head>
    <style>
      body {
        font-family: Arial, sans-serif;
        background-color: #f4f4f4;
        color: #333;
        margin: 0;
        padding: 0;
      }
      .container {
        max-width: 600px;
        margin: 50px auto;
        background-color: #ffffff;
        padding: 20px;
        border-radius: 8px;
        box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
      }
      h1 {
        color: #333;
        text-align: center;
      }
      p {
        font-size: 16px;
        line-height: 1.5;
      }
      code {
        display: block;
        background-color: #f0f0f0;
        color: #ff8c00;
        padding: 10px;
        font-size: 14px;
        margin: 20px auto;
        border-radius: 4px;
        width: 90%;
        word-wrap: break-word;
      }
      .footer {
        text-align: center;
        font-size: 12px;
        color: #aaa;
        margin-top: 20px;
      }
    </style>
  </head>
  <body>
    <div class="container">
      <h1>Hello!</h1>
      <p>
        Use the following command in your CLI to reset your password:
      </p>
      <code> syftbox reset-password --email {{ email }} --token {{ token }} </code>
      <p>If you did not request this, please ignore this email.</p>
    </div>
  </body>
</html>
"""


def send_token_email(server_settings, user_email: str, token: str):
    template = Template(token_email_template)
    body = template.render(email=user_email, token=token)
    send_email(
        server_settings=server_settings,
        receiver_email=user_email,
        subject="SyftBox Token",
        body=body,
        mimetype="text/html",
    )


def send_email(
    server_settings: ServerSettings,
    receiver_email: str,
    subject: str,
    body: str,
    mimetype: str = "text/html",
):
    payload = {
        "personalizations": [{"to": [{"email": receiver_email}]}],
        "from": {"email": SENDER_EMAIL},
        "subject": subject,
        "content": [{"type": mimetype, "value": body}],
    }

    headers = {
        "Authorization": f"Bearer {server_settings.sendgrid_secret.get_secret_value()}",
        "Content-Type": "application/json",
    }

    try:
        response = httpx.post(SENDGRID_SERVER, json=payload, headers=headers, timeout=10.0)
        response.raise_for_status()
        logger.info(f"Email sent to {receiver_email}")
        return {"success": True, "status_code": response.status_code}
    except httpx.HTTPError as e:
        logger.error(str(e))



from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from syftbox.lib.constants import DEFAULT_CONFIG_PATH

__all__ = ["syftbox_env"]


class SyftEnvVars(BaseSettings):
    """SyftBox environment variables."""

    DISABLE_ICONS: bool = Field(default=False)
    """Disable copying icons to the datasite dir."""

    CLIENT_CONFIG_PATH: Path = Field(default=DEFAULT_CONFIG_PATH)
    """Path to the client configuration file."""

    model_config = SettingsConfigDict(env_file="client.env", env_prefix="SYFTBOX_")


syftbox_env = SyftEnvVars()



from pathlib import Path

from syftbox.client.base import SyftClientInterface
from syftbox.client.plugins.sync.sync import DatasiteState
from syftbox.client.utils.dir_tree import create_dir_tree
from syftbox.client.utils.display import display_file_tree
from syftbox.lib.ignore import IGNORE_FILENAME, filter_ignored_paths

ignore_file = """
# Exlude alice datasite
/alice@example.com

# exclude all occurrences bob@example.com
bob@example.com

# Exclude all "large" folders under any datasite
*/large/*

# Include important_file.pdf under excluded folder
!/john@example.com/large/important_file.pdf

# General excludes
*.tmp
_.syftignore
*.py[cod]
"""

paths_with_result = [
    # Should be ignored
    ("alice@example.com/file1.txt", True),
    ("john@example.com/results/bob@example.com/file1.txt", True),
    ("john@example.com/large/file1.txt", True),
    ("john@example.com/docs/file1.tmp", True),
    ("script.pyc", True),
    # Should not be ignored
    ("john@example.com/results/alice@example.com/file1.txt", False),
    ("john@example.com/large/important_file.pdf", False),
    ("john@example.com/docs/file3.pdf", False),
    ("script.py", False),
]


def test_ignore_file(tmp_path):
    # without ignore file
    ignore_path = tmp_path / IGNORE_FILENAME
    ignore_path.unlink(missing_ok=True)

    paths, results = zip(*paths_with_result)
    paths = [Path(p) for p in paths]
    filtered_paths = filter_ignored_paths(tmp_path, paths)
    assert filtered_paths == paths

    # with ignore file
    ignore_path.write_text(ignore_file)

    expected_result = [p for p, r in zip(paths, results) if r is False]
    filtered_paths = filter_ignored_paths(tmp_path, paths)
    assert filtered_paths == expected_result


def test_ignore_datasite(datasite_1: SyftClientInterface, datasite_2: SyftClientInterface) -> None:
    datasite_2_files = {
        datasite_2.email: {
            "visible_file.txt": "content",
            "ignored_file.pyc": "content",
        }
    }
    num_files = 2
    num_visible_files = 1
    create_dir_tree(Path(datasite_1.workspace.datasites), datasite_2_files)
    display_file_tree(Path(datasite_1.workspace.datasites))

    # ds1 gets their local state of ds2
    datasite_state = DatasiteState(client=datasite_1, email=datasite_2.email)
    _, local_changes = datasite_state.get_out_of_sync_files()

    assert len(local_changes) == num_visible_files
    assert local_changes[0].path == Path(datasite_2.email) / "visible_file.txt"

    # ds1 ignores ds2
    ignore_path = Path(datasite_1.workspace.datasites) / IGNORE_FILENAME
    with ignore_path.open("a") as f:
        # /datasite_2/
        f.write(f"\n/{datasite_2.email}\n")

    # ds1 gets their local state of ds2
    _, local_changes = datasite_state.get_out_of_sync_files()
    assert len(local_changes) == 0

    # remove ignore file
    ignore_path.unlink()
    _, local_changes = datasite_state.get_out_of_sync_files()
    assert len(local_changes) == num_files


def test_ignore_symlinks(datasite_1: SyftClientInterface) -> None:
    # create a symlinked folder containing a file
    folder_to_symlink = Path(datasite_1.workspace.data_dir) / "folder"
    folder_to_symlink.mkdir()
    symlinked_file = folder_to_symlink / "file.txt"
    symlinked_file.write_text("content")

    symlinked_folder = Path(datasite_1.workspace.datasites) / "symlinked_folder"
    symlinked_folder.symlink_to(folder_to_symlink)

    paths = [
        Path("symlinked_folder/file.txt"),
        Path("symlinked_folder/non_existent_file.txt"),
        Path("symlinked_folder/subfolder/file.txt"),
    ]

    filtered_paths = filter_ignored_paths(datasite_1.workspace.datasites, paths, ignore_symlinks=True)
    assert filtered_paths == []


def test_ignore_hidden_files(datasite_1: SyftClientInterface) -> None:
    paths = [
        Path(".hidden_file.txt"),  # Hidden files are filtered
        Path("visible_file.txt"),  # Visible files are not filtered
        Path("subfolder/.hidden_file.txt"),  # Hidden files in folders are filtered
        Path(".subfolder/visible_file.txt"),  # Files in hidden folders are filtered
    ]

    filtered_paths = filter_ignored_paths(datasite_1.workspace.datasites, paths, ignore_hidden_files=True)
    assert filtered_paths == [Path("visible_file.txt")]



from collections.abc import Generator
from functools import partial
from pathlib import Path

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from syftbox.client.base import SyftClientInterface
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.datasite import create_datasite
from syftbox.lib.workspace import SyftWorkspace
from syftbox.server.server import app as server_app
from syftbox.server.server import lifespan as server_lifespan
from syftbox.server.settings import ServerSettings
from tests.unit.server.conftest import get_access_token


class MockClient(SyftClientInterface):
    def __init__(self, config, workspace, server_client):
        self.config = config
        self.workspace = workspace
        self.server_client = server_client

        access_token = get_access_token(self.server_client, self.email)
        self.server_client.headers["Authorization"] = f"Bearer {access_token}"

    @property
    def email(self):
        return self.config.email

    @property
    def datasite(self):
        return Path(self.workspace.datasites, self.config.email)

    @property
    def all_datasites(self) -> list[str]:
        """List all datasites in the workspace"""
        return [d.name for d in self.workspace.datasites.iterdir() if (d.is_dir() and "@" in d.name)]


def setup_datasite(tmp_path: Path, server_client: TestClient, email: str) -> SyftClientInterface:
    syft_path = tmp_path / email
    config = SyftClientConfig(
        path=syft_path / "config.json",
        data_dir=syft_path,
        email=email,
        server_url=str(server_client.base_url),
        client_url="http://localhost:8080",
    )
    config.save()
    ws = SyftWorkspace(config.data_dir)
    ws.mkdirs()
    create_datasite(ws.datasites, email)
    return MockClient(config, ws, server_client)


@pytest.fixture(scope="function")
def server_app_with_lifespan(tmp_path: Path) -> FastAPI:
    """
    NOTE we are spawning a new server thread for each datasite,
    this is not ideal but it is the same as using multiple uvicorn workers
    """
    path = tmp_path / "server"
    path.mkdir()

    settings = ServerSettings.from_data_folder(path)
    lifespan_with_settings = partial(server_lifespan, settings=settings)
    server_app.router.lifespan_context = lifespan_with_settings

    return server_app


@pytest.fixture()
def datasite_1(tmp_path: Path, server_app_with_lifespan: FastAPI) -> SyftClientInterface:
    email = "user_1@openmined.org"
    with TestClient(server_app_with_lifespan) as client:
        return setup_datasite(tmp_path, client, email)


@pytest.fixture()
def datasite_2(tmp_path: Path, server_app_with_lifespan: FastAPI) -> SyftClientInterface:
    email = "user_2@openmined.org"
    with TestClient(server_app_with_lifespan) as client:
        return setup_datasite(tmp_path, client, email)


@pytest.fixture(scope="function")
def server_client(server_app_with_lifespan: FastAPI) -> Generator[TestClient, None, None]:
    with TestClient(server_app_with_lifespan) as client:
        yield client



import hashlib
import json
import os
import shutil
import subprocess
import threading
import time
from datetime import datetime
from pathlib import Path
from types import SimpleNamespace

from croniter import croniter
from loguru import logger
from typing_extensions import Any, Optional, Union

from syftbox.client.base import SyftClientInterface
from syftbox.lib.client_config import CONFIG_PATH_ENV

APP_LOG_FILE_NAME_FORMAT = "{app_name}.log"
DEFAULT_INTERVAL = 10
RUNNING_APPS = {}
DEFAULT_APPS_PATH = Path(os.path.join(os.path.dirname(__file__), "..", "..", "..", "default_apps")).absolute().resolve()
EVENT = threading.Event()


def path_without_virtualenvs() -> str:
    env_path = os.getenv("PATH", "")
    if not env_path:
        return env_path

    venv_hints = [
        f"env{os.sep}bin",
        f"env{os.sep}Scripts",
        "conda",
        ".virtualenvs",
        "pyenv",
    ]

    # activated venv will have VIRTUAL_ENV and VIRTUAL_ENV/bin in PATH
    # so axe it
    env_venv = os.getenv("VIRTUAL_ENV", "")
    if env_venv:
        venv_hints.append(env_venv)

    cleaned_path = [
        entry for entry in env_path.split(os.pathsep) if not any(hint in entry.lower() for hint in venv_hints)
    ]

    return os.pathsep.join(cleaned_path)


def get_clean_env():
    clean_env = {}

    essential_vars = {
        "PATH",
        "HOME",
        "USER",
        "TEMP",
        "TMP",
        "TMPDIR",
        "SHELL",
        "LANG",
        "LC_ALL",
        "DISPLAY",  # X11 specific (Linux)
        "DBUS_SESSION_BUS_ADDRESS",  # X11 specific (Linux)
        "SYSTEMROOT",  # Windows specific
    }

    # Copy essential and SYFTBOX_* variables
    for key, value in os.environ.items():
        if key in essential_vars or key.startswith("SYFTBOX_"):
            clean_env[key] = value

    return clean_env


def find_and_run_script(app_path: Path, extra_args: list, config_path: Path, app_log_dir: Path):
    script_path = os.path.join(app_path, "run.sh")

    clean_env = get_clean_env()
    clean_env.update(
        {
            "PATH": path_without_virtualenvs(),
            CONFIG_PATH_ENV: str(config_path),
        }
    )

    # Check if the script exists
    if os.path.isfile(script_path):
        # Set execution bit (+x)
        os.chmod(script_path, os.stat(script_path).st_mode | 0o111)

        # Prepare the command based on whether there's a shebang or not
        command = ["sh", script_path] + extra_args

        result, _ = run_with_logging(
            command,
            app_path,
            clean_env,
            app_log_dir,
        )
        return result
    else:
        raise FileNotFoundError(f"run.sh not found in {app_path}")


def create_app_logger(log_file: Path):
    """Create an isolated logger for app runs"""
    import logging
    from logging.handlers import RotatingFileHandler

    # Create a new logger instance
    logger = logging.getLogger(f"app_logger_{log_file.name}")
    logger.setLevel(logging.INFO)

    # Remove any existing handlers
    logger.handlers.clear()

    # Create formatter
    formatter = logging.Formatter("%(asctime)s | %(levelname)-8s | %(message)s", datefmt="%Y-%m-%d at %H:%M:%S")

    # Create and configure file handler
    file_handler = RotatingFileHandler(
        filename=log_file,
        maxBytes=100 * 1024 * 1024,  # 100Mb
        backupCount=3,
        encoding="utf-8",
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.INFO)

    # Add handler to logger
    logger.addHandler(file_handler)

    return logger, file_handler


def run_with_logging(command: str, app_path: Path, clean_env: dict, log_path: Path):
    """
    Run a subprocess command and capture output to both a log file and return results.
    """
    # Create logs directory if it doesn't exist
    log_path.mkdir(parents=True, exist_ok=True)

    # Create a unique log filename with timestamp and app name
    app_name = app_path.name
    log_file = log_path / APP_LOG_FILE_NAME_FORMAT.format(app_name=app_name)

    # Create isolated logger for this run
    app_logger, file_handler = create_app_logger(log_file=log_file)

    try:
        # Log run metadata
        app_logger.info(f"Working directory: {app_path}")
        app_logger.info(f"Command: {command}")

        # Run the subprocess
        process = subprocess.run(
            command,
            cwd=app_path,
            check=True,
            capture_output=True,
            text=True,
            env=clean_env,
        )

        # log this in app log
        app_logger.info(
            (
                f"exit code: {process.returncode}\n"
                f"===stdout===\n{process.stdout}"
                f"===stderr===\n{process.stderr or '-'}"
            )
        )
        # log this in console
        logger.info(f"Process completed with exit code: {process.returncode}. Log file: {log_file}")
        return process, log_file

    except subprocess.CalledProcessError as e:
        # log this in app log
        app_logger.error(
            ("process output\n" f"> exit code: {e.returncode}\n" f"> stdout:\n{e.stdout}" f"> stderr:\n{e.stderr}")
        )
        # log this in console
        logger.error(f"Process failed with exit code: {e.returncode}")
        raise e

    except Exception as e:
        app_logger.error(f"Unexpected error: {str(e)}")
        raise e

    finally:
        app_logger.removeHandler(file_handler)
        file_handler.close()


def copy_default_apps(apps_path: Path):
    if not DEFAULT_APPS_PATH.exists():
        logger.info(f"Default apps directory not found: {DEFAULT_APPS_PATH}")
        return

    for app in DEFAULT_APPS_PATH.iterdir():
        src_app_path = DEFAULT_APPS_PATH / app
        dst_app_path = apps_path / app.name

        if src_app_path.is_dir():
            if dst_app_path.exists():
                logger.info(f"App already installed at: {dst_app_path}")
                # shutil.rmtree(dst_app_path)
            else:
                shutil.copytree(src_app_path, dst_app_path)
                logger.info(f"Copied default app:: {app}")


def dict_to_namespace(data) -> Union[SimpleNamespace, list, Any]:
    if isinstance(data, dict):
        return SimpleNamespace(**{key: dict_to_namespace(value) for key, value in data.items()})
    elif isinstance(data, list):
        return [dict_to_namespace(item) for item in data]
    else:
        return data


def load_config(path: str) -> Optional[SimpleNamespace]:
    try:
        with open(path, "r") as f:
            data = json.load(f)
        return dict_to_namespace(data)
    except Exception:
        return None


def bootstrap(client: SyftClientInterface):
    # create the directory
    apps_path = client.workspace.apps

    apps_path.mkdir(exist_ok=True)

    # Copy default apps if they don't exist
    copy_default_apps(apps_path)


def run_apps(apps_path: Path, client_config: Path):
    # create the directory

    for app in apps_path.iterdir():
        app_path = apps_path.absolute() / app
        if app_path.is_dir():
            app_config = load_config(app_path / "config.json")
            if app_config is None:
                run_app(app_path, client_config)
            elif RUNNING_APPS.get(app, None) is None:
                logger.info("⏱  Scheduling a  new app run.")
                thread = threading.Thread(
                    target=run_custom_app_config,
                    args=(
                        app_config,
                        app_path,
                        client_config,
                    ),
                )
                thread.start()
                RUNNING_APPS[os.path.basename(app)] = thread


def get_file_hash(file_path, digest="md5") -> str:
    with open(file_path, "rb") as f:
        return hashlib.file_digest(f, digest)


def output_published(app_output, published_output) -> bool:
    return (
        os.path.exists(app_output)
        and os.path.exists(published_output)
        and get_file_hash(app_output, "md5") == get_file_hash(published_output, "md5")
    )


def run_custom_app_config(app_config: SimpleNamespace, app_path: Path, client_config: Path):
    app_name = os.path.basename(app_path)
    clean_env = {
        "PATH": path_without_virtualenvs(),
        CONFIG_PATH_ENV: str(client_config),
    }
    # Update environment with any custom variables in app_config
    app_envs = getattr(app_config.app, "env", {})
    if not isinstance(app_envs, dict):
        app_envs = vars(app_envs)
    clean_env.update(app_envs)

    # Retrieve the cron-style schedule from app_config
    cron_iter = None
    interval = None
    cron_schedule = getattr(app_config.app.run, "schedule", None)
    if cron_schedule is not None:
        base_time = datetime.now()
        cron_iter = croniter(cron_schedule, base_time)
    elif getattr(app_config.app.run, "interval", None) is not None:
        interval = app_config.app.run.interval
    else:
        raise Exception("There's no schedule configuration. Please add schedule or interval in your app config.json")

    while not EVENT.is_set():
        current_time = datetime.now()
        logger.info(f"👟 Running {app_name} at scheduled time {current_time.strftime('%Y-%m-%d %H:%M:%S')}")
        logger.info(f"Running command: {app_config.app.run.command}")
        try:
            app_log_dir = app_path / "logs"
            run_with_logging(
                app_config.app.run.command,
                app_path,
                clean_env,
                app_log_dir,
            )
            log_file = app_log_dir / APP_LOG_FILE_NAME_FORMAT.format(app_name=app_name)
            logger.info(f"App '{app_name}' ran successfully. \nDetailed logs at: {log_file.resolve()}")
        except subprocess.CalledProcessError as _:
            logger.error(f"Error calling subprocess for api '{app_name}'")
            logger.error(f"Check {app_name}'s api logs at: {log_file.resolve()}")
        except Exception as _:
            logger.error(f"Error running '{app_name}'")
            logger.error(f"Check {app_name} api logs at: {log_file.resolve()}")

        if cron_iter is not None:
            # Schedule the next execution
            next_execution = cron_iter.get_next(datetime)
            time_to_wait = int((next_execution - current_time).total_seconds())
            logger.info(
                f"⏲ Waiting for scheduled time. Current time: {current_time.strftime('%Y-%m-%d %H:%M:%S')}, Next execution: {next_execution.strftime('%Y-%m-%d %H:%M:%S')}"
            )
        else:
            time_to_wait = int(interval)
        time.sleep(time_to_wait)


def run_app(app_path: Path, config_path: Path):
    app_name = os.path.basename(app_path)
    app_log_dir = app_path / "logs"
    log_file = app_log_dir / APP_LOG_FILE_NAME_FORMAT.format(app_name=app_name)

    extra_args = []
    try:
        logger.info(f"Running '{app_name}' app")
        find_and_run_script(app_path, extra_args, config_path, app_log_dir)
        logger.info(f"`{app_name}` App ran successfully. \nDetailed logs at: {log_file.resolve()}")
    except FileNotFoundError as e:
        logger.error(f"Error running '{app_name}'")
        logger.error(f"Error: {str(e)}")
    except subprocess.CalledProcessError as _:
        logger.error(f"Error calling subprocess for api '{app_name}'")
        logger.error(f"Check {app_name}'s api logs at: {log_file.resolve()}")
    except Exception as _:
        logger.error(f"Error running '{app_name}'")
        logger.error(f"Check {app_name} api logs at: {log_file.resolve()}")


class AppRunner:
    def __init__(self, client: SyftClientInterface, interval: int = DEFAULT_INTERVAL):
        self.client = client
        self.__event = threading.Event()
        self.interval = interval
        self.__run_thread: threading.Thread

    def start(self):
        def run():
            bootstrap(self.client)
            while not self.__event.is_set():
                try:
                    run_apps(
                        apps_path=self.client.workspace.apps,
                        client_config=self.client.config.path,
                    )
                    time.sleep(self.interval)
                except Exception as e:
                    logger.error(f"Error running apps: {str(e)}")

        self.__run_thread = threading.Thread(target=run)
        self.__run_thread.start()

    def stop(self, blocking: bool = False):
        if not self.__run_thread:
            return

        EVENT.set()
        self.__event.set()
        blocking and self.__run_thread.join()



"""
SyftBox CLI - Setup scripts
"""

import json
import shutil
from pathlib import Path

from rich import print as rprint
from rich.prompt import Confirm, Prompt

from syftbox.__version__ import __version__
from syftbox.client.auth import authenticate_user
from syftbox.client.client2 import METADATA_FILENAME
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.constants import DEFAULT_DATA_DIR
from syftbox.lib.exceptions import ClientConfigException
from syftbox.lib.validators import DIR_NOT_EMPTY, is_valid_dir, is_valid_email

__all__ = ["setup_config_interactive"]


def is_empty(data_dir: Path) -> bool:
    """True if the data_dir is empty"""
    return not any(data_dir.iterdir())


def has_old_syftbox_version(data_dir: Path) -> bool:
    """True if the data_dir was created with an older version of SyftBox"""
    metadata_file = data_dir / METADATA_FILENAME
    if not metadata_file.exists():
        return True
    metadata = json.loads(metadata_file.read_text())
    current_version = __version__
    old_version = metadata.get("version", None)
    return old_version != current_version


def prompt_delete_old_data_dir(data_dir: Path) -> bool:
    msg = f"[yellow]Found old SyftBox folder at {data_dir}.[/yellow]\n"
    msg += "[yellow]Press Y to remove the old folder and download it from the server [bold](recommended)[/bold]. Press N to keep the old folder and migrate it.[/yellow]"
    return Confirm.ask(msg)


def get_migration_decision(data_dir: Path):
    migrate_datasite = False
    if data_dir.exists():
        if is_empty(data_dir):
            migrate_datasite = False
        elif has_old_syftbox_version(data_dir):
            # we need this extra if because we do 2 things:
            # 1. determine if we want to remove
            # 2. determine if we want to migrate
            if prompt_delete_old_data_dir(data_dir):
                rprint("Removing old syftbox folder")
                shutil.rmtree(str(data_dir))
                migrate_datasite = False
            else:
                migrate_datasite = True
    return migrate_datasite


def setup_config_interactive(
    config_path: Path, email: str, data_dir: Path, server: str, port: int, skip_auth: bool = False
) -> SyftClientConfig:
    """Setup the client configuration interactively. Called from CLI"""

    config_path = config_path.expanduser().resolve()
    conf: SyftClientConfig = None
    if data_dir:
        data_dir = data_dir.expanduser().resolve()

    # try to load the existing config
    try:
        conf = SyftClientConfig.load(config_path)
    except ClientConfigException:
        pass

    if not conf:
        # first time setup
        if not data_dir or data_dir == DEFAULT_DATA_DIR:
            data_dir = prompt_data_dir()

        if not email:
            email = prompt_email()

        # create a new config with the input params
        conf = SyftClientConfig(
            path=config_path,
            sync_folder=data_dir,
            email=email,
            server_url=server,
            port=port,
        )
    else:
        if server and server != conf.server_url:
            conf.set_server_url(server)
        if port != conf.client_url.port:
            conf.set_port(port)

    if not skip_auth:
        conf.access_token = authenticate_user(conf)

    # DO NOT SAVE THE CONFIG HERE.
    # We don't know if the client will accept the config yet
    return conf


def prompt_data_dir(default_dir: Path = DEFAULT_DATA_DIR) -> Path:
    prompt_dir = "[bold]Where do you want SyftBox to store data?[/bold] [grey70]Press Enter for default[/grey70]"
    prompt_overwrite = "[bold yellow]Directory '{sync_folder}' is not empty![/bold yellow] Do you want to overwrite it?"

    while True:
        sync_folder = Prompt.ask(prompt_dir, default=str(default_dir))
        valid, reason = is_valid_dir(sync_folder)
        if reason == DIR_NOT_EMPTY:
            overwrite = Confirm.ask(prompt_overwrite.format(sync_folder=sync_folder))
            if not overwrite:
                continue
            valid = True

        if not valid:
            rprint(f"[bold red]{reason}[/bold red] '{sync_folder}'")
            continue

        path = Path(sync_folder).expanduser().resolve()
        rprint(f"Selected directory [bold]'{path}'[/bold]")
        return path


def prompt_email() -> str:
    while True:
        email = Prompt.ask("[bold]Enter your email address[/bold]")
        if not is_valid_email(email):
            rprint(f"[bold red]Invalid email[/bold red]: '{email}'")
            continue
        return email



import threading
from dataclasses import dataclass
from pathlib import Path
from queue import PriorityQueue
from typing import Dict, Optional

from syftbox.client.plugins.sync.sync import FileChangeInfo


@dataclass(order=True)
class SyncQueueItem:
    priority: int
    data: FileChangeInfo


class SyncQueue:
    """
    A thread-safe priority queue that supports deduplication based on the data field.

    Adding an item to the queue that already exists will be ignored, even if the priority is different.
    """

    def __init__(self, maxsize: int = 0):
        self.queue: PriorityQueue[SyncQueueItem] = PriorityQueue(maxsize=maxsize)
        self.all_items: Dict[Path, SyncQueueItem] = {}

        self.lock = threading.Lock()

    def put(self, item: SyncQueueItem, block: bool = False, timeout: Optional[float] = None) -> None:
        with self.lock:
            if item.data.path not in self.all_items:
                self.queue.put(item, block=block, timeout=timeout)
                self.all_items[item.data.path] = item

    def get(self, block: bool = False, timeout: Optional[float] = None) -> SyncQueueItem:
        with self.lock:
            item: SyncQueueItem = self.queue.get(block=block, timeout=timeout)
            self.all_items.pop(item.data.path, None)
            return item

    def empty(self) -> bool:
        return self.queue.empty()



# routers/datasite_router.py

from fastapi import APIRouter, HTTPException
from loguru import logger
from pydantic import BaseModel

from syftbox.client.routers.common import APIContext

router = APIRouter()


# Don't think we require this Request model until we have
# an endpoint that allows one to create a datasite
class DatasiteRequest(BaseModel):
    name: str


@router.get("/")
async def list_datasites(ctx: APIContext):
    """List all available datasites"""

    try:
        return {"datasites": ctx.all_datasites}
    except Exception as e:
        logger.error(f"Error listing datasites: {e}")
        raise HTTPException(status_code=500, detail="Failed to list datasites")



import time
from threading import Thread
from typing import Optional

from loguru import logger

from syftbox.client.base import SyftClientInterface
from syftbox.client.exceptions import SyftAuthenticationError
from syftbox.client.plugins.sync.consumer import SyncConsumer
from syftbox.client.plugins.sync.endpoints import get_datasite_states, whoami
from syftbox.client.plugins.sync.exceptions import FatalSyncError
from syftbox.client.plugins.sync.queue import SyncQueue, SyncQueueItem
from syftbox.client.plugins.sync.sync import DatasiteState, FileChangeInfo


class SyncManager:
    def __init__(self, client: SyftClientInterface, health_check_interval: int = 300):
        self.client = client
        self.queue = SyncQueue()
        self.consumer = SyncConsumer(client=self.client, queue=self.queue)
        self.sync_interval = 1  # seconds
        self.thread: Optional[Thread] = None
        self.is_stop_requested = False
        self.sync_run_once = False

        self.last_health_check = 0
        self.health_check_interval = health_check_interval

    def is_alive(self) -> bool:
        return self.thread is not None and self.thread.is_alive()

    def stop(self, blocking: bool = False):
        self.is_stop_requested = True
        if blocking:
            self.thread.join()

    def start(self):
        def _start(manager: SyncManager):
            while not manager.is_stop_requested:
                try:
                    if manager._should_perform_health_check():
                        manager.check_server_sync_status()
                    manager.run_single_thread()
                    time.sleep(manager.sync_interval)
                except FatalSyncError as e:
                    logger.error(f"Syncing encountered a fatal error. {e}")
                    break

        self.is_stop_requested = False
        t = Thread(target=_start, args=(self,), daemon=True)
        t.start()
        logger.info(f"Sync started, syncing every {self.sync_interval} seconds")
        self.thread = t

    def enqueue(self, change: FileChangeInfo) -> None:
        self.queue.put(SyncQueueItem(priority=change.get_priority(), data=change))

    def get_datasite_states(self) -> list[DatasiteState]:
        try:
            remote_datasite_states = get_datasite_states(self.client.server_client, email=self.client.email)
        except Exception as e:
            logger.error(f"Failed to retrieve datasites from server, only syncing own datasite. Reason: {e}")
            remote_datasite_states = {}

        # Ensure we are always syncing own datasite
        if self.client.email not in remote_datasite_states:
            remote_datasite_states[self.client.email] = []

        datasite_states = [
            DatasiteState(self.client, email, remote_state=remote_state)
            for email, remote_state in remote_datasite_states.items()
        ]
        return datasite_states

    def _should_perform_health_check(self) -> bool:
        return time.time() - self.last_health_check > self.health_check_interval

    def check_server_sync_status(self):
        """
        check if the server is still available for syncing,
        if the user cannot authenticate, the sync will stop.

        Raises:
            FatalSyncError: If the server is not available.
        """
        try:
            _ = whoami(self.client.server_client)
            logger.debug("Health check succeeded, server is available.")
            self.last_health_check = time.time()
        except SyftAuthenticationError as e:
            # Auth errors will never recover, sync should be stopped
            raise FatalSyncError(f"Health check failed, {e}")
        except Exception as e:
            logger.error(f"Health check failed: {e}. Retrying in {self.health_check_interval} seconds.")

    def enqueue_datasite_changes(self, datasite: DatasiteState):
        try:
            permission_changes, file_changes = datasite.get_out_of_sync_files()
            total = len(permission_changes) + len(file_changes)

            if total != 0:
                logger.debug(
                    f"Enqueuing {len(permission_changes)} permissions and {len(file_changes)} files for {datasite.email}"
                )
        except Exception as e:
            logger.error(f"Failed to get out of sync files for {datasite.email}. Reason: {e}")
            permission_changes, file_changes = [], []

        for change in permission_changes + file_changes:
            self.enqueue(change)

    def run_single_thread(self):
        # NOTE first implementation will be unthreaded and just loop through all datasites

        datasite_states = self.get_datasite_states()
        logger.debug(f"Syncing {len(datasite_states)} datasites")

        if not self.sync_run_once:
            # Download all missing files at the start
            self.consumer.download_all_missing(
                datasite_states=datasite_states,
            )

        for datasite_state in datasite_states:
            self.enqueue_datasite_changes(datasite_state)

        # TODO stop consumer if self.is_stop_requested
        self.consumer.consume_all()

        self.sync_run_once = True



# /// script
# dependencies = [
#    "opendp==0.11.1",
#    "syftbox==0.1.0",
#    "pandas==2.2.3",
# ]
#
# [tool.uv.sources]
# syftbox = { path = "/Users/madhavajay/dev/syft", editable = true }
# ///

__name__ = "myanalysis"
__author__ = "david.rolle@gmail.com"


def input_reader(private: bool = False):
    import pandas as pd

    from syftbox.lib import sy_path

    inputs = {}
    inputs["trade_data"] = pd.read_csv(sy_path("./inputs/trade_data/trade_mock.csv", resolve_private=private))
    return inputs


def output_writer(result, private: bool = False):
    import json

    output_path = "./output/result/result.json"
    if not private:
        output_path = output_path.replace(".json", ".mock.json")
    with open(output_path, "w") as f:
        f.write(json.dumps(result))


# START YOUR CODE


def myanalysis(trade_data):
    import opendp.prelude as dp

    dp.enable_features("contrib")

    aggregate = 0.0
    base_lap = dp.m.make_laplace(
        dp.atom_domain(T=float),
        dp.absolute_distance(T=float),
        scale=5.0,
    )
    noise = base_lap(aggregate)

    total = trade_data["Trade Value (US$)"].sum()
    result = (float(total / 1_000_000), float(noise))

    print(result)
    if result[0] > 3:
        print("Got mock")
    else:
        print("Got private")
    return result


# END YOUR CODE


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Process some input.")
    parser.add_argument("--private", action="store_true", help="Run in private mode")
    args = parser.parse_args()

    print(f"Running: {__name__} from {__author__}")
    inputs = input_reader(private=args.private)
    print("> Reading Inputs", inputs)

    output = myanalysis(**inputs)

    print("> Writing Outputs", output)
    output_writer(output, private=args.private)
    print(f"> ✅ Running {__name__} Complete!")


main()



import json
import os
import platform
import re
import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path
from tempfile import mkdtemp
from types import SimpleNamespace

from typing_extensions import Any, Optional


def is_git_installed() -> bool:
    """
    Checks if Git is installed on the system.

    Returns:
        bool: `True` if Git is installed, `False` otherwise.

    Functionality:
        - Runs the `git --version` command to check if Git is installed.
        - If the command runs successfully, returns `True`.
        - If the command fails (e.g., Git is not installed), returns `False`.

    Example:
        ```python
        if is_git_installed():
            print("Git is installed on this system.")
        else:
            print("Git is not installed. Please install Git to proceed.")
        ```
        This will print a message indicating whether Git is installed or not.
    """
    try:
        subprocess.run(
            ["git", "--version"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def sanitize_git_path(path: str) -> str:
    """
    Validates and sanitizes a Git repository path, ensuring it matches the required format.

    Args:
        path (str): The Git repository path to validate.

    Returns:
        str: The sanitized Git repository path if it matches the valid pattern.

    Raises:
        ValueError: If the provided path does not match the expected format for a Git repository.

    Functionality:
        - Uses a regular expression pattern to ensure that the given path follows the format `owner/repository`.
        - If the path matches the pattern, returns it as a valid Git path.
        - If the path does not match the pattern, raises a `ValueError` with a descriptive message.

    Example:
        Suppose you have a GitHub path like `OpenMined/logged_in` and want to validate it:
        ```python
        try:
            sanitized_path = sanitize_git_path("OpenMined/logged_in")
        except ValueError as e:
            print(e)
        ```
        If the path is valid, `sanitized_path` will contain the validated GitHub path. If it is not valid, the error message
        "Invalid Git repository path format. (eg: OpenMined/logged_in)" will be printed.
    """

    if path.startswith("http://"):
        path = path.replace("http://", "")

    if path.startswith("https://"):
        path = path.replace("https://", "")

    if path.startswith("github.com/"):
        path = path.replace("github.com/", "")

    # Define a regex pattern for a valid GitHub path
    pattern = r"^[a-zA-Z0-9_.-]+/[a-zA-Z0-9_.-]+$"

    # Check if the path matches the pattern
    if re.match(pattern, path):
        return path
    else:
        raise ValueError("Invalid Git repository path format. (eg: OpenMined/logged_in)")


def delete_folder_if_exists(folder_path: str) -> None:
    """
    Deletes a folder if it exists at the specified path.

    Args:
        folder_path (str): The path to the folder to be deleted.

    Returns:
        None: This function does not return any value.

    Functionality:
        - Checks if the folder exists at the given path.
        - If the folder exists and is a directory, deletes it and all of its contents using `shutil.rmtree()`.

    Example:
        Suppose you want to delete a folder located at `/tmp/old_clone` if it exists:
        ```python
        delete_folder_if_exists("/tmp/old_clone")
        ```
        This will delete the folder and all of its contents if it exists.
    """
    if os.path.exists(folder_path) and os.path.isdir(folder_path):
        shutil.rmtree(folder_path)


def is_repo_accessible(repo_url: str) -> bool:
    """
    Checks if the specified Git repository is accessible.

    Args:
        repo_url (str): The URL of the Git repository to check.

    Returns:
        bool: `True` if the repository is accessible, `False` otherwise.

    Functionality:
        - Uses the `git ls-remote` command to check if the Git repository is accessible.
        - If the command succeeds, returns `True`.
        - If the command fails or times out, returns `False`.

    Example:
        Suppose you want to check if a repository located at `https://github.com/example/repo.git` is accessible.
        You can call the function like this:
        ```python
        is_accessible = is_repo_accessible("https://github.com/example/repo.git")
        ```
        This will return `True` if the repository is accessible, or `False` if it is not.
    """
    try:
        env = os.environ.copy()
        env["GIT_TERMINAL_PROMPT"] = "0"
        subprocess.run(
            ["git", "ls-remote", repo_url],
            check=True,
            env=env,
            capture_output=True,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def clone_repository(sanitized_git_path: str, branch: str) -> str:
    """
    Clones a Git repository from GitHub to a temporary directory.

    Args:
        sanitized_git_path (str): The Git repository path in the format `owner/repository`.

    Returns:
        str: The path to the cloned repository.

    Raises:
        Exception: If Git is not installed on the system.
        ValueError: If the provided repository path is not accessible.
        CalledProcessError: If there is an error during the cloning process.

    Functionality:
        - Checks if Git is installed on the system by calling `is_git_installed()`.
        - Forms the GitHub repository URL from the provided `sanitized_git_path`.
        - Checks if the repository is accessible by calling `is_repo_accessible()`.
        - Clones the repository to a temporary directory (`/tmp`).
        - Deletes any existing folder in `/tmp` with the same name before cloning.
        - If cloning is successful, returns the path to the cloned repository.
        - If any error occurs during cloning, raises the corresponding exception.

    Example:
        Suppose you want to clone a repository located at `OpenMined/PySyft` to a temporary directory.
        You can call the function like this:
        ```python
        try:
            clone_path = clone_repository("OpenMined/PySyft")
            print(f"Repository cloned to: {clone_path}")
        except Exception as e:
            print(e)
        ```
        This will clone the repository to `/tmp/PySyft` if successful, or print an error message if any issues occur.
    """
    if not is_git_installed():
        raise Exception(
            "git cli isn't installed. Please, follow the instructions"
            + " to install git according to your OS. (eg. brew install git)"
        )
    repo_url = f"https://github.com/{sanitized_git_path}.git"
    if not is_repo_accessible(repo_url):
        raise ValueError(f"Cannot access repository {repo_url}")

    # Clone repository in /tmp
    tmp_path = mkdtemp(prefix="syftbox_app_")
    temp_clone_path = Path(tmp_path, sanitized_git_path.split("/")[-1])

    # Delete if there's already an existent repository folder in /tmp path.
    delete_folder_if_exists(temp_clone_path)

    try:
        subprocess.run(
            [
                "git",
                "clone",
                "-b",
                branch,
                "--single-branch",
                repo_url,
                temp_clone_path,
            ],
            check=True,
            text=True,
            capture_output=True,
        )
        return temp_clone_path
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr)


def dict_to_namespace(data: Any) -> Any:
    """
    Converts a dictionary (or nested dictionary) to a SimpleNamespace object.

    Args:
        data (dict or list): The data to convert. Can be a dictionary, list of dictionaries, or other types.

    Returns:
        SimpleNamespace or list: A SimpleNamespace object representing the dictionary data,
                                or a list of SimpleNamespace objects if the input is a list.
                                If the input is not a dictionary or list, returns the input as-is.

    Functionality:
        - Recursively converts dictionaries to SimpleNamespace objects.
        - If the data is a list, each item in the list is recursively converted.
        - If the data is neither a dictionary nor a list, returns the data unchanged.

    Example:
        Suppose you have a dictionary with nested data:
        ```python
        data = {
            "user": {
                "name": "Alice",
                "age": 30,
                "address": {
                    "city": "Wonderland",
                    "zipcode": "12345"
                }
            },
            "active": True
        }
        namespace_data = dict_to_namespace(data)
        print(namespace_data.user.name)  # Output: Alice
        print(namespace_data.user.address.city)  # Output: Wonderland
        ```
        This will allow you to access dictionary values using dot notation like attributes.
    """
    if isinstance(data, dict):
        return SimpleNamespace(**{key: dict_to_namespace(value) for key, value in data.items()})
    elif isinstance(data, list):
        return [dict_to_namespace(item) for item in data]
    else:
        return data


def load_config(path: str) -> SimpleNamespace:
    """
    Loads a JSON configuration file and converts it to a SimpleNamespace object.

    Args:
        path (str): The file path to the JSON configuration file.

    Returns:
        SimpleNamespace: A SimpleNamespace object representing the configuration data.

    Raises:
        ValueError: If the file does not exist, is not in JSON format, or does not contain a dictionary.

    Functionality:
        - Checks if the provided file path exists. If not, raises a `ValueError` indicating the file is not found.
        - Opens and reads the JSON file. If the file cannot be decoded or does not contain a dictionary, raises a `ValueError`.
        - Converts the loaded dictionary to a SimpleNamespace object for easy attribute-based access.

    Example:
        Suppose you have a JSON configuration file at `/path/to/config.json` with the following content:
        ```json
        {
            "version": "0.1.0",
            "app": {
                "version": "1.0"
                "env": {
                    "TEST_ENV": "testing",
                },
            },
        }
        ```
        You can load the configuration and access its fields using dot notation:
        ```python
        try:
            config = load_config("/path/to/config.json")
            print(config.app.version)  # Output: MyApp
            print(config.app.env.TEST_ENV)  # Output: True
        except ValueError as e:
            print(e)
        ```
        This will load the configuration and allow access to its values using attribute access.
    """
    if not os.path.exists(path):
        raise ValueError(f"config not found - {path}")
    try:
        error_msg = "File isn't in JSON format."
        with open(path, "r") as f:
            data = json.load(f)
        if not isinstance(data, dict):
            raise ValueError(error_msg)
    except json.JSONDecodeError:
        raise ValueError(error_msg)
    return dict_to_namespace(data)


def create_symbolic_link(apps_dir: Path, sanitized_path: str):
    """
    Creates a symbolic link from the application directory in the Syftbox directory to the user's sync folder.

    Args:
        apps_dir (Path): The path to the `apps` directory in the Syftbox configuration folder.
        app_path (str): The actual path of the application directory.
        sanitized_path (str): The sanitized Git repository path in the format `owner/repository`.

    Returns:
        None: This function does not return any value.

    Functionality:
        - Constructs the symbolic link path within the user's sync folder (`apps` folder).
        - If a symlink already exists at the target location, deletes it to avoid conflicts.
        - Creates a new symbolic link pointing from the sync folder to the application directory.

    Example:
        Suppose you want to create a symbolic link for an application located at `/home/user/.syftbox/apps/PySyft`:
        ```python
        create_symbolic_link(
            apps_dir=SyftWorkspace.apps, # ex "/home/user/SyftBox/apis",
            sanitized_path="OpenMined/PySyft"
        )
        ```
        This will create a symbolic link at `<sync_folder>/apps/PySyft` pointing to the application directory.
    """
    # TODO: Create a Symlink function
    # - Handles if path doesn't exists.
    target_symlink_path = f"{apps_dir}/{sanitized_path.split('/')[-1]}"

    # Create the symlink
    if os.path.exists(target_symlink_path) and os.path.islink(target_symlink_path):
        os.unlink(target_symlink_path)

    if not os.path.exists(target_symlink_path):
        os.symlink(sanitized_path, target_symlink_path)
    else:
        raise Exception(f"Path exists and isn't a symlink: {target_symlink_path}")
    return target_symlink_path


def move_repository_to_syftbox(apps_dir: Path, tmp_clone_path: str, sanitized_path: str) -> str:
    """
    Moves a cloned Git repository to the Syftbox directory.

    Args:
        tmp_clone_path (str): The file path to the temporarily cloned Git repository.
        sanitized_path (str): The sanitized Git repository path in the format `owner/repository`.

    Returns:
        str: The final destination path of the moved repository.

    Functionality:
        - Constructs the destination path within the Syftbox configuration directory (`apps` folder).
        - Deletes any existing folder at the destination path to avoid conflicts.
        - Moves the repository from the temporary clone path to the destination path.
        - Returns the new path of the moved repository.

    Example:
        Suppose you have cloned a repository to a temporary path `/tmp/syftbox` and want to move it to the Syftbox directory:
        ```python
        output_path = move_repository_to_syftbox("/tmp/PySyft", "OpenMined/PySyft")
        print(output_path)  # Output: /path/to/config/apps/PySyft
        ```
        This will move the cloned repository to the Syftbox `apps` directory and return the final destination path.
    """
    output_path = f"{apps_dir}/{sanitized_path.split('/')[-1]}"
    delete_folder_if_exists(output_path)
    shutil.move(tmp_clone_path, output_path)
    return output_path


def run_pre_install(app_config: SimpleNamespace, app_path: str):
    """
    Runs pre-installation commands specified in the application configuration.

    Args:
        app_config (SimpleNamespace): The configuration object for the application, which is expected to have an `app`
                                      attribute with a `pre_install` attribute containing a list of commands to run.
        app_path (string): The file path to the app folder.

    Returns:
        None: This function does not return any value.

    Functionality:
        - Checks if the `pre_install` attribute exists and contains commands in the application configuration.
        - If the `pre_install` attribute is empty or does not exist, the function returns without executing any command.
        - If there are pre-installation commands, runs them using `subprocess.run()`.

    Example:
        Suppose you have an application configuration that specifies a pre-installation command to install dependencies:
        ```python
        app_config = SimpleNamespace(
            app=SimpleNamespace(pre_install=["echo", "Installing dependencies..."])
        )
        run_pre_install(app_config)
        ```
        This will run the specified pre-installation command using `subprocess.run()`.
    """
    if len(getattr(app_config.app, "pre_install", [])) == 0:
        return

    try:
        subprocess.run(
            app_config.app.pre_install,
            cwd=app_path,
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr)


def run_post_install(app_config: SimpleNamespace, app_path: str):
    """
    Runs post-installation commands specified in the application configuration.

    Args:
        app_config (SimpleNamespace): The configuration object for the application, which is expected to have an `app`
                                      attribute with a `post_install` attribute containing a list of commands to run.

    Returns:
        None: This function does not return any value.

    Functionality:
        - Checks if the `post_install` attribute exists and contains commands in the application configuration.
        - If the `post_install` attribute is empty or does not exist, the function returns without executing any command.
        - If there are post-installation commands, runs them using `subprocess.run()`.

    Example:
        Suppose you have an application configuration that specifies a post-installation command to perform cleanup:
        ```python
        app_config = SimpleNamespace(
            app=SimpleNamespace(post_install=["echo", "Performing post-installation cleanup..."])
        )
        run_post_install(app_config)
        ```
        This will run the specified post-installation command using `subprocess.run()`.
    """
    if len(getattr(app_config.app, "post_install", [])) == 0:
        return

    try:
        subprocess.run(
            app_config.app.post_install,
            cwd=app_path,
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr)


def check_os_compatibility(app_config) -> None:
    """
    Checks whether the current operating system is compatible with the application based on the configuration.

    Args:
        app_config: The configuration object for the application, which is expected to have an `app` attribute
                    with a `platforms` attribute containing a list of supported operating systems.

    Returns:
        None: This function does not return any value.

    Raises:
        OSError: If the current operating system is not supported by the application.

    Functionality:
        - Uses the `platform.system()` function to determine the current operating system.
        - Checks the application's configuration (`app_config`) for a list of supported operating systems.
        - If no platforms are defined in the configuration, the function simply returns without doing anything.
        - If the current operating system is not in the list of supported platforms, raises an `OSError`.

    Example:
        Suppose you have an application configuration that specifies supported platforms as `['Windows', 'Linux']`.
        The function will determine the current operating system and raise an `OSError` if it is not supported:
        ```python
        try:
            check_os_compatibility(app_config)
        except OSError as e:
            print(e)
        ```
        If the current OS is not in the supported platforms list, the message "Your OS isn't supported by this app." will be printed.
    """
    os_name = platform.system().lower()
    supported_os = getattr(app_config.app, "platforms", [])

    # If there's no platforms field in config.json, just ignore it.
    if len(supported_os) == 0:
        return

    is_compatible = False
    for operational_system in supported_os:
        if operational_system.lower() == os_name:
            is_compatible = True

    if not is_compatible:
        raise OSError("Your OS isn't supported by this app.")


def get_current_commit(app_path: str) -> str:
    """
    Retrieves the current commit hash for a Git repository located at the specified path.

    Args:
        app_path (str): The file path to the Git repository.

    Returns:
        str: The current commit hash of the repository if the command is successful.
             If an error occurs, returns an error message describing the failure.

    Functionality:
        - Uses the `git rev-parse HEAD` command to get the current commit hash.
        - If the command succeeds, returns the commit hash as a string.
        - If the command fails (e.g., if the provided path is not a valid Git repository),
          returns an error message detailing what went wrong.

    Example:
        Suppose you have a Git repository at `/path/to/repo` and want to retrieve its current commit hash.
        You can call the function like this:
        ```python
        commit_hash = get_current_commit("/path/to/repo")
        ```
        This will return the commit hash if the repository exists and the command runs successfully,
        or an error message if there is an issue with the command.
    """
    try:
        # Navigate to the repository path and get the current commit hash
        commit_hash = (
            subprocess.check_output(["git", "-C", app_path, "rev-parse", "HEAD"], stderr=subprocess.STDOUT)
            .strip()
            .decode("utf-8")
        )
        return commit_hash
    except subprocess.CalledProcessError:
        return "local"


def update_app_config_file(app_path: str, sanitized_git_path: str, app_config) -> None:
    """
    Updates the `app.json` configuration file with the current commit and version information of an application.

    Args:
        app_path (str): The file path of the application.
        sanitized_git_path (str): The sanitized path representing the Git repository.
        app_config: The configuration object for the application, which is expected to have an `app` attribute
                    with a `version` attribute, if available.

    Returns:
        None: This function modifies the `app.json` configuration file in place and returns nothing.

    Functionality:
        - Normalizes the provided application path.
        - Determines the configuration directory by navigating two levels up from the application path.
        - Checks if an `app.json` file exists in the configuration directory.
            - If it exists, loads its contents into a dictionary.
            - If it does not exist, creates an empty dictionary for new configuration entries.
        - Retrieves the current commit information of the application using the `get_current_commit` function.
        - If the application version is available from the `app_config` object, includes it in the configuration.
        - Updates the `app.json` configuration file with the new commit and version information under the key
          specified by `sanitized_git_path`.
        - Writes the updated configuration back to the `app.json` file with indentation for readability.

    Example:
        Suppose you have an application located at `/path/to/app` and you want to update the `app.json` file
        with the latest commit and version. You can call the function like this:
        ```python
        update_app_config_file("/path/to/app", "my_sanitized_git_path", app_config)
        ```
        This will update or create entries in `app.json` for the given Git path, storing commit and version details.
    """
    normalized_app_path = os.path.normpath(app_path)

    conf_path = os.path.dirname(os.path.dirname(normalized_app_path))

    app_json_path = conf_path + "/app.json"
    app_json_config = {}

    if os.path.exists(app_json_path):
        # Read from it.
        with open(app_json_path, "r") as app_json_file:
            app_json_config = json.load(app_json_file)

    app_version = None
    if getattr(app_config.app, "version", None) is not None:
        app_version = app_config.app.version

    current_commit = get_current_commit(normalized_app_path)
    if current_commit == "local":
        app_version = "dev"

    app_json_config[sanitized_git_path] = {
        "commit": current_commit,
        "version": app_version,
        "path": app_path,
    }

    with open(app_json_path, "w") as json_file:
        json.dump(app_json_config, json_file, indent=4)


def check_app_config(tmp_clone_path) -> Optional[SimpleNamespace]:
    app_config_path = Path(tmp_clone_path) / "config.json"
    if os.path.exists(app_config_path):
        app_config = load_config(app_config_path)
        check_os_compatibility(app_config)
        return app_config
    return None


@dataclass
class InstallResult:
    app_name: str
    app_path: Path
    error: Optional[Exception]
    details: Optional[str]


def install(apps_dir: Path, repository: str, branch: str) -> InstallResult:
    """
    Installs an application by cloning the repository, checking compatibility, and running installation scripts.

    Args:
        apps_dir (Path): Path where app will be installed.

    Returns:
        None: If the installation is successful.
        Tuple[str, Exception]: If an error occurs during any installation step, returns a tuple with the step description and the exception raised.

    Functionality:
        - Parses command-line arguments to get the Git repository to install.
        - Performs a series of steps to install the application, including:
            1. Sanitizing the Git repository path.
            2. Cloning the repository to a temporary directory.
            3. Loading the application's configuration (`config.json`).
            4. Checking platform compatibility.
            5. Moving the repository to the Syftbox directory.
            6. Creating a symbolic link on the user's desktop.
            7. Running pre-installation commands.
            8. Running post-installation commands.
            9. Updating the `apps.json` file to include the installed application.
        - If any step fails, returns the step description and the exception raised.

    Example:
        Suppose you have a client configuration and want to install an application from a repository:
        ```python
        result = install(Path("~/.syftbox/apps"), "OpenMined/PySyft", "main")
        if result.error:
            print(f"Error installing {result.app_name}: {result.error}")
            print(f"Failed at step: {result.details}")
        else:
            print(f"Successfully installed {result.app_name} at {result.app_path}")
        ```
        This will install the application, and if an error occurs, it will indicate the step where the failure happened.
    """
    step = ""
    try:
        # NOTE:
        # Sanitize git repository path
        # Handles: bad format repository path.
        # Returns: Sanitized repository path.
        step = "checking app name"

        sanitized_path = repository
        if not os.path.exists(repository):
            sanitized_path = sanitize_git_path(repository)

            # NOTE:
            # Clones the app repository
            # Handles: Git cli tool not installed.
            # Handles: Repository path doesn't exits / isn't public.
            # Handles: If /tmp/apps/<repository_name> already exists (replaces it)
            # Returns: Path where the repository folder was cloned temporarily.
            step = "pulling App"
            tmp_clone_path = clone_repository(sanitized_path, branch)

            # NOTE:
            # Load config.json
            # Handles: config.json doesn't exist in the pulled repository
            # Handles: config.json version is different from syftbox config version.
            # Returns: Loaded app config as SimpleNamespace instance.
        else:
            tmp_clone_path = os.path.abspath(repository)

        # make optional
        app_config = None
        try:
            check_app_config(tmp_clone_path)
        except Exception:
            # this function is run in cli context
            # dont loguru here, either rprint or bubble up the error
            app_config = None

        # NOTE:
        # Moves the repository from /tmp to ~/.syftbox/apps/<repository_name>
        # Handles: If ~/.syftbox/apps/<repository_name> already exists (replaces it)
        if not os.path.exists(repository):
            app_config_path = move_repository_to_syftbox(
                apps_dir,
                tmp_clone_path=tmp_clone_path,
                sanitized_path=sanitized_path,
            )
        else:
            # Creates a Symbolic Link ( ~/Desktop/Syftbox/app/<rep> -> ~/.syftbox/apps/<rep>)
            # Handles: If ~/.syftbox/apps/<repository_name> already exists (replaces it)
            step = "creating Symbolic Link"
            app_config_path = create_symbolic_link(
                apps_dir=apps_dir,
                sanitized_path=tmp_clone_path,
            )

        # NOTE:
        # Executes config.json pre-install command list
        # Handles: Exceptions from pre-install command execution
        if app_config:
            step = "running pre-install commands"
            run_pre_install(app_config, app_config_path)

        # NOTE:
        # Executes config.json post-install command list
        # Handles: Exceptions from post-install command execution
        if app_config:
            step = "running post-install commands"
            run_post_install(app_config, app_config_path)

        # NOTE:
        # Updates the apps.json file
        # Handles: If apps.json file doesn't exist yet.
        # Handles: If apps.json already have the repository_name  app listed.
        # Handles: If apps.json exists but doesn't have the repository_name app listed.
        if app_config:
            step = "updating apps.json config"
            update_app_config_file(app_config_path, sanitized_path, app_config)

        app_dir = Path(app_config_path)
        return InstallResult(app_name=app_dir.name, app_path=app_dir, error=None, details=None)
    except Exception as e:
        return InstallResult(app_name="", app_path=Path(""), error=e, details=step)



from pathlib import Path
from subprocess import CalledProcessError

import pytest

from syftbox.app.install import clone_repository, sanitize_git_path


def test_valid_git_path():
    path = "Example/Repository"
    output_path = sanitize_git_path(path)
    assert path == output_path


def test_valid_git_url():
    path = "Example/Repository"
    http_url = f"http://github.com/{path}"
    output_path = sanitize_git_path(http_url)
    assert path == output_path

    https_url = f"https://github.com/{path}"
    output_path = sanitize_git_path(https_url)
    assert path == output_path


def test_invalid_git_path():
    path = "..Example/../Repository"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_second_invalid_git_path():
    path = "http://example.com"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_clone_valid_repository(monkeypatch):
    count = 0

    def mock_subproc_run(*args, **kwargs):
        nonlocal count
        count += 1
        return 0

    monkeypatch.setattr("subprocess.run", mock_subproc_run)

    path = "OpenMined/logged_in"
    temp_path = clone_repository(path, "main")
    assert count == 3
    assert isinstance(temp_path, Path)


def test_clone_repository_to_an_existent_path(monkeypatch):
    count = 0

    def mock_subproc_run(*args, **kwargs):
        nonlocal count
        count += 1
        return 0

    monkeypatch.setattr("subprocess.run", mock_subproc_run)

    # First call will make the repository path exist
    path = "OpenMined/logged_in"
    temp_path = clone_repository(path, "main")
    assert isinstance(temp_path, Path)
    assert count == 3

    # Second call must clone it again without any exception (replaces the old one).
    temp_path = clone_repository(path, "main")
    assert isinstance(temp_path, Path)
    assert count == 6


def test_clone_invalid_repository(monkeypatch):
    count = 0

    def mock_subproc_run(*args, **kwargs):
        nonlocal count
        count += 1
        cmd = args[0]

        if cmd[0] == "git" and cmd[1] == "ls-remote" and "Invalid" in cmd[2]:
            raise CalledProcessError(1, cmd)

        return 0

    monkeypatch.setattr("subprocess.run", mock_subproc_run)

    path = "InvalidUser/InvalidRepo"
    with pytest.raises(ValueError) as excpt:
        _ = clone_repository(path, "main")
        assert "Cannot access repository" in excpt.value



import random
from pathlib import Path
from queue import Empty

import pytest
from pydantic import BaseModel

from syftbox.client.plugins.sync.queue import SyncQueue, SyncQueueItem


class MockFileChangeInfo(BaseModel):  # noqa: F821
    path: Path


def test_sync_queue():
    queue = SyncQueue()

    n = 10

    paths = [Path(f"file_{i}.txt") for i in range(n)]
    priorities = [random.uniform(0, 1000) for _ in range(n)]
    priorities[0] = int(priorities[0])  # int and float should both work
    items = [SyncQueueItem(priority, MockFileChangeInfo(path=path)) for path, priority in zip(paths, priorities)]
    items_sorted = sorted(items, key=lambda x: x.priority)

    for item in items:
        queue.put(item)

    assert not queue.empty()
    assert set(queue.all_items.keys()) == set(paths)

    for item in items_sorted:
        assert queue.get() == item

    assert queue.empty()
    assert len(queue.all_items) == 0
    with pytest.raises(Empty):
        queue.get(block=False)


def test_sync_queue_dedupe():
    queue = SyncQueue()

    path = Path("file.txt")

    queue.put(SyncQueueItem(1, MockFileChangeInfo(path=path)))
    assert set(queue.all_items.keys()) == {path}
    assert not queue.empty()

    for _ in range(10):
        queue.put(SyncQueueItem(random.random(), MockFileChangeInfo(path=path)))

    assert set(queue.all_items.keys()) == {path}

    queue.get()
    assert len(queue.all_items) == 0
    assert queue.empty()



import os
import platform
import shutil
import sys
from pathlib import PurePath
from typing import Any, Optional
from venv import logger

import psutil

from syftbox.__version__ import __version__
from syftbox.app.manager import list_app
from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.types import PathLike
from syftbox.lib.workspace import SyftWorkspace


def debug_report(config_path: Optional[PathLike] = None) -> str:
    client_config = None
    apps = []
    app_dir = None
    try:
        client_config = SyftClientConfig.load(config_path)
        workspace = SyftWorkspace(client_config.data_dir)
        result = list_app(workspace)
        app_dir = result.apps_dir
        apps = [app.name for app in result.apps]
        client_config = client_config.as_dict(exclude={"token", "access_token"})
    except Exception as e:
        logger.exception("Error loading client config", e)
        pass

    syftbox_path = shutil.which("syftbox")

    return {
        "system": {
            "resources": {
                "cpus": psutil.cpu_count(logical=True),
                "architecture": platform.machine(),
                "ram": f"{psutil.virtual_memory().total / (1024**3):.2f} GB",
            },
            "operating_system": {
                "name": "macOS" if platform.system() == "Darwin" else platform.system(),
                "version": platform.release(),
            },
            "python": {
                "version": platform.python_version(),
                "binary_location": sys.executable,
            },
        },
        "syftbox": {
            "version": __version__,
            "command": syftbox_path or "syftbox executable not found in PATH",
            "client_config_path": config_path,
            "client_config": client_config,
            "apps_dir": app_dir,
            "apps": apps,
        },
        "syftbox_env": {key: value for key, value in os.environ.items() if key.startswith("SYFT")},
    }


def debug_report_yaml(config_path: Optional[PathLike] = None) -> str:
    import yaml
    from pydantic_core import Url

    def str_representer(dumper: yaml.Dumper, val: Any) -> yaml.ScalarNode:
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(val))

    # Register the custom representers
    yaml.add_multi_representer(PurePath, str_representer)
    yaml.add_representer(Url, str_representer)

    report = debug_report(config_path)
    return yaml.dump(report, default_flow_style=False, sort_keys=False)
